import copy
from typing import List, Optional, Dict

from antlr.xDroneParser import xDroneParser
from antlr.xDroneParserVisitor import xDroneParserVisitor
from xdrone.compiler.compiler_utils.drones import NullDrone, Drone
from xdrone.compiler.compiler_utils.expressions import Identifier, ListElem, VectorElem, Expression
from xdrone.compiler.compiler_utils.functions import FunctionTable, Function, Parameter, FunctionIdentifier
from xdrone.compiler.compiler_utils.symbol_table import SymbolTable
from xdrone.compiler.compiler_utils.type import Type, ListType, EmptyList
from xdrone.shared.command import Command, SingleDroneCommand, ParallelDroneCommands, AbstractDroneCommand, \
    RepeatDroneNameException
from xdrone.shared.compile_error import CompileError
from xdrone.shared.drone_config import DroneConfig


class Compiler(xDroneParserVisitor):

    def __init__(self, drone_config_map: Dict[str, DroneConfig],
                 symbol_table: SymbolTable, function_table: FunctionTable):
        super().__init__()

        self.drones = {name: Drone(name, config) for name, config in drone_config_map.items()}  # global constant table
        self.symbol_table = [symbol_table]
        self.function_table = function_table

        self.returned = [False]
        self.returned_value = []

        self.commands = [[]]

    def _get_latest_symbol_table(self):
        return self.symbol_table[-1]

    def _get_latest_commands(self):
        return self.commands[-1]

    ######## prog ########

    def visitProg(self, ctx: xDroneParser.ProgContext) -> List[AbstractDroneCommand]:
        for func in ctx.func():
            self.visit(func)
        self.visit(ctx.commands())
        assert len(self.commands) == 1
        commands = self.commands[0]
        return commands

    ######## commands ########

    def visitCommands(self, ctx: xDroneParser.CommandsContext) -> None:
        for command in ctx.command():
            if self.returned[-1]:
                break
            self.visit(command)

    ######## command ########

    def _get_drone_name(self, expr: Optional[Expression]):
        if expr is not None:
            if expr.type != Type.drone():
                raise CompileError("Expression {} should have type drone, but is {}".format(expr, expr.type))
            if isinstance(expr.value, NullDrone):
                raise CompileError("Drone has not been assigned".format(expr, expr.type))

            drone_name = expr.value.name
        elif expr is None and len(self.drones) == 1:
            drone_name = list(self.drones.keys())[0]
        else:
            raise CompileError("Drone should be specified if there are multiple drones in config")
        return drone_name

    def visitTakeoff(self, ctx: xDroneParser.TakeoffContext) -> None:
        expr = self.visit(ctx.expr()) if ctx.expr() else None
        drone_name = self._get_drone_name(expr)
        self._get_latest_commands().append(SingleDroneCommand(drone_name, Command.takeoff()))

    def visitLand(self, ctx: xDroneParser.LandContext) -> None:
        exprs = self.visit(ctx.expr()) if ctx.expr() else None
        drone_name = self._get_drone_name(exprs)
        self._get_latest_commands().append(SingleDroneCommand(drone_name, Command.land()))

    def visitUp(self, ctx: xDroneParser.UpContext) -> None:
        exprs = [self.visit(expr) for expr in ctx.expr()]
        if ctx.DOT():
            drone_expr, expr = exprs
        else:
            drone_expr, expr = None, exprs[0]
        if expr.type != Type.int() and expr.type != Type.decimal():
            raise CompileError("Expression {} should have type int or decimal, but is {}".format(expr, expr.type))
        drone_name = self._get_drone_name(drone_expr)
        self._get_latest_commands().append(SingleDroneCommand(drone_name, Command.up(expr.value)))

    def visitDown(self, ctx: xDroneParser.DownContext) -> None:
        exprs = [self.visit(expr) for expr in ctx.expr()]
        if ctx.DOT():
            drone_expr, expr = exprs
        else:
            drone_expr, expr = None, exprs[0]
        if expr.type != Type.int() and expr.type != Type.decimal():
            raise CompileError("Expression {} should have type int or decimal, but is {}".format(expr, expr.type))
        drone_name = self._get_drone_name(drone_expr)
        self._get_latest_commands().append(SingleDroneCommand(drone_name, Command.down(expr.value)))

    def visitLeft(self, ctx: xDroneParser.LeftContext) -> None:
        exprs = [self.visit(expr) for expr in ctx.expr()]
        if ctx.DOT():
            drone_expr, expr = exprs
        else:
            drone_expr, expr = None, exprs[0]
        if expr.type != Type.int() and expr.type != Type.decimal():
            raise CompileError("Expression {} should have type int or decimal, but is {}".format(expr, expr.type))
        drone_name = self._get_drone_name(drone_expr)
        self._get_latest_commands().append(SingleDroneCommand(drone_name, Command.left(expr.value)))

    def visitRight(self, ctx: xDroneParser.RightContext) -> None:
        exprs = [self.visit(expr) for expr in ctx.expr()]
        if ctx.DOT():
            drone_expr, expr = exprs
        else:
            drone_expr, expr = None, exprs[0]
        if expr.type != Type.int() and expr.type != Type.decimal():
            raise CompileError("Expression {} should have type int or decimal, but is {}".format(expr, expr.type))
        drone_name = self._get_drone_name(drone_expr)
        self._get_latest_commands().append(SingleDroneCommand(drone_name, Command.right(expr.value)))

    def visitForward(self, ctx: xDroneParser.ForwardContext) -> None:
        exprs = [self.visit(expr) for expr in ctx.expr()]
        if ctx.DOT():
            drone_expr, expr = exprs
        else:
            drone_expr, expr = None, exprs[0]
        if expr.type != Type.int() and expr.type != Type.decimal():
            raise CompileError("Expression {} should have type int or decimal, but is {}".format(expr, expr.type))
        drone_name = self._get_drone_name(drone_expr)
        self._get_latest_commands().append(SingleDroneCommand(drone_name, Command.forward(expr.value)))

    def visitBackward(self, ctx: xDroneParser.BackwardContext) -> None:
        exprs = [self.visit(expr) for expr in ctx.expr()]
        if ctx.DOT():
            drone_expr, expr = exprs
        else:
            drone_expr, expr = None, exprs[0]
        if expr.type != Type.int() and expr.type != Type.decimal():
            raise CompileError("Expression {} should have type int or decimal, but is {}".format(expr, expr.type))
        drone_name = self._get_drone_name(drone_expr)
        self._get_latest_commands().append(SingleDroneCommand(drone_name, Command.backward(expr.value)))

    def visitRotateLeft(self, ctx: xDroneParser.RotateLeftContext) -> None:
        exprs = [self.visit(expr) for expr in ctx.expr()]
        if ctx.DOT():
            drone_expr, expr = exprs
        else:
            drone_expr, expr = None, exprs[0]
        if expr.type != Type.int() and expr.type != Type.decimal():
            raise CompileError("Expression {} should have type int or decimal, but is {}".format(expr, expr.type))
        drone_name = self._get_drone_name(drone_expr)
        self._get_latest_commands().append(SingleDroneCommand(drone_name, Command.rotate_left(expr.value)))

    def visitRotateRight(self, ctx: xDroneParser.RotateRightContext) -> None:
        exprs = [self.visit(expr) for expr in ctx.expr()]
        if ctx.DOT():
            drone_expr, expr = exprs
        else:
            drone_expr, expr = None, exprs[0]
        if expr.type != Type.int() and expr.type != Type.decimal():
            raise CompileError("Expression {} should have type int or decimal, but is {}".format(expr, expr.type))
        drone_name = self._get_drone_name(drone_expr)
        self._get_latest_commands().append(SingleDroneCommand(drone_name, Command.rotate_right(expr.value)))

    def visitWait(self, ctx: xDroneParser.WaitContext) -> None:
        exprs = [self.visit(expr) for expr in ctx.expr()]
        if ctx.DOT():
            drone_expr, expr = exprs
        else:
            drone_expr, expr = None, exprs[0]
        if expr.type != Type.int() and expr.type != Type.decimal():
            raise CompileError("Expression {} should have type int or decimal, but is {}".format(expr, expr.type))
        drone_name = self._get_drone_name(drone_expr)
        self._get_latest_commands().append(SingleDroneCommand(drone_name, Command.wait(expr.value)))

    def visitDeclare(self, ctx: xDroneParser.DeclareContext) -> None:
        type, identifier = self.visit(ctx.type_()), self.visit(ctx.ident())
        ident = identifier.ident
        if ident in self._get_latest_symbol_table() or ident in self.drones:
            raise CompileError("Identifier {} already declared".format(ident))
        self._get_latest_symbol_table().store(ident, Expression(type, type.default_value, ident=ident))

    def visitDeclareAssign(self, ctx: xDroneParser.DeclareAssignContext) -> None:
        type, identifier, expr = self.visit(ctx.type_()), self.visit(ctx.ident()), self.visit(ctx.expr())
        ident = identifier.ident
        if ident in self._get_latest_symbol_table() or identifier.ident in self.drones:
            raise CompileError("Identifier {} already declared".format(ident))
        if expr.type != type:
            raise CompileError("Identifier {} has been declared as {}, but assigned as {}"
                               .format(ident, type, expr.type))
        expr_with_ident = Expression(type, expr.value, ident)
        self._get_latest_symbol_table().store(ident, expr_with_ident)

    def _unfold_nested_list(self, ident: str) -> (str, list, list):
        if "[" in ident:
            tokens = ident.split("[")
            ident = tokens[0]
            indices = [int(token.replace("]", "")) for token in tokens[1:]]
        else:
            indices = []
        assert ident in self._get_latest_symbol_table()
        new_list = self._get_latest_symbol_table().get_expression(ident).value
        inner = new_list
        for i in indices:
            inner = inner[i]
        return ident, new_list, inner

    def _insert_nested_ident(self, ident: Optional[str], expr: Expression, index: int) -> None:
        if ident is not None:
            new_ident, new_list, inner = self._unfold_nested_list(ident)
            inner.insert(index, expr.value)
            self._get_latest_symbol_table().update(new_ident, new_list)

    def _remove_nested_ident(self, ident: Optional[str], index: int) -> None:
        if ident is not None:
            new_ident, new_list, inner = self._unfold_nested_list(ident)
            inner.pop(index)
            self._get_latest_symbol_table().update(new_ident, new_list)

    def _update_nested_ident(self, ident: Optional[str], expr: Expression, index: int) -> None:
        if ident is not None:
            new_ident, new_list, inner = self._unfold_nested_list(ident)
            inner[index] = expr.value
            self._get_latest_symbol_table().update(new_ident, new_list)

    def visitAssignVectorElem(self, ctx: xDroneParser.AssignVectorElemContext) -> None:
        vector_elem, expr = self.visit(ctx.vectorElem()), self.visit(ctx.expr())
        ident = vector_elem.ident
        vector = vector_elem.container
        index = vector_elem.index
        if expr.type != Type.int() and expr.type != Type.decimal():
            raise CompileError("Assigned value {} should have type int or decimal, but is {}".format(expr, expr.type))
        decimal_expr = Expression(Type.decimal(), float(expr.value), ident=expr.ident)
        self._update_nested_ident(ident, decimal_expr, index)

    def visitAssignListElem(self, ctx: xDroneParser.AssignListElemContext) -> None:
        list_elem, expr = self.visit(ctx.listElem()), self.visit(ctx.expr())
        ident = list_elem.ident
        list = list_elem.container
        index = list_elem.index
        assigned_type = expr.type
        declared_type = list.type.elem_type
        if assigned_type != declared_type:
            raise CompileError("Assigned value {} should have type {}, but is {}"
                               .format(expr, declared_type, assigned_type))
        self._update_nested_ident(ident, expr, index)

    def visitAssignIdent(self, ctx: xDroneParser.AssignIdentContext) -> None:
        identifier, expr = self.visit(ctx.ident()), self.visit(ctx.expr())
        ident = identifier.ident
        if identifier.ident in self.drones:
            raise CompileError("Identifier {} is a drone constant, cannot be assigned".format(ident))
        if ident not in self._get_latest_symbol_table():
            raise CompileError("Identifier {} has not been declared".format(ident))
        assigned_type = expr.type
        declared_type = self._get_latest_symbol_table().get_expression(ident).type
        if assigned_type != declared_type:
            raise CompileError("Identifier {} has been declared as {}, but assigned as {}"
                               .format(ident, declared_type, assigned_type))
        self._get_latest_symbol_table().update(ident, expr.value)

    def visitDel(self, ctx: xDroneParser.DelContext) -> None:
        identifier = self.visit(ctx.ident())
        ident = identifier.ident
        if identifier.ident in self.drones:
            raise CompileError("Identifier {} is a drone constant, cannot be deleted".format(ident))
        if ident not in self._get_latest_symbol_table():
            raise CompileError("Identifier {} has not been declared".format(ident))
        self._get_latest_symbol_table().delete(ident)

    def visitInsert(self, ctx: xDroneParser.InsertContext) -> None:
        list = self.visit(ctx.expr(0))
        if not isinstance(list.type, ListType):
            raise CompileError("Expression {} should have type list, but is {}".format(list, list.type))
        if ctx.AT():
            index = self.visit(ctx.expr(1))
            value = self.visit(ctx.expr(2))
        else:
            index = Expression(Type.int(), len(list.value))
            value = self.visit(ctx.expr(1))
        if index.type != Type.int():
            raise CompileError("Expression {} should have type int, but is {}".format(index, index.type))
        if index.value > len(list.value) or index.value < 0:
            raise CompileError("List {} has length {}, but has been inserted at out-of-range index {}"
                               .format(list, len(list.value), index.value))
        if not isinstance(list.type, EmptyList) and value.type != list.type.elem_type:
            raise CompileError("List {} has been declared as {}, but inserted with element type {}"
                               .format(list, list.type, value.type))
        self._insert_nested_ident(list.ident, value, index.value)

    def visitRemove(self, ctx: xDroneParser.RemoveContext) -> None:
        list = self.visit(ctx.expr(0))
        if not isinstance(list.type, ListType):
            raise CompileError("Expression {} should have type list, but is {}".format(list, list.type))
        if ctx.AT():
            index = self.visit(ctx.expr(1))
        else:
            index = Expression(Type.int(), len(list.value) - 1)
        if index.type != Type.int():
            raise CompileError("Expression {} should have type int, but is {}".format(index, index.type))
        if index.value >= len(list.value) or index.value < 0:
            raise CompileError("List {} has length {}, but has been removed at out-of-range index {}"
                               .format(list, len(list.value), index.value))
        self._remove_nested_ident(list.ident, index.value)

    def visitProcedureCall(self, ctx: xDroneParser.ProcedureCallContext) -> None:
        call = self.visit(ctx.call())
        if call is not None:
            raise CompileError("Procedure call should not return any expression, but {} is returned".format(call))
        return call

    def _visit_commands_with_scope(self, commands_ctx):
        # keep updates on existing variables, discard new variables
        prev_idents = self._get_latest_symbol_table().idents()
        self.visit(commands_ctx)
        self._get_latest_symbol_table().keep_idents(prev_idents)

    def visitIf(self, ctx: xDroneParser.IfContext) -> None:
        expr = self.visit(ctx.expr())
        if expr.type != Type.boolean():
            raise CompileError("Expression {} should have type boolean, but is {}".format(expr, expr.type))
        if expr.value:
            # scope - keep updates on existing variables, discard new variables
            self._visit_commands_with_scope(ctx.commands(0))
        else:
            if ctx.ELSE():
                # scope - keep updates on existing variables, discard new variables
                self._visit_commands_with_scope(ctx.commands(1))

    def visitWhile(self, ctx: xDroneParser.WhileContext) -> None:
        expr = self.visit(ctx.expr())
        if expr.type != Type.boolean():
            raise CompileError("Expression {} should have type boolean, but is {}".format(expr, expr.type))
        while expr.value:
            # scope - keep updates on existing variables, discard new variables
            self._visit_commands_with_scope(ctx.commands())
            expr = self.visit(ctx.expr())

    def visitFor(self, ctx: xDroneParser.ForContext) -> None:
        identifier, expr1, expr2 = self.visit(ctx.ident()), self.visit(ctx.expr(0)), self.visit(ctx.expr(1))
        ident = identifier.ident
        if ident not in self._get_latest_symbol_table() and ident not in self.drones:
            raise CompileError("Identifier {} has not been declared".format(ident))
        declared_type = self._get_latest_symbol_table().get_expression(ident).type
        if declared_type != Type.int():
            raise CompileError("Identifier {} has been declared as {}, but assigned as {}"
                               .format(ident, declared_type, Type.int()))
        if expr1.type != Type.int():
            raise CompileError("Expression {} should have type int, but is {}".format(expr1, expr1.type))
        if expr2.type != Type.int():
            raise CompileError("Expression {} should have type int, but is {}".format(expr2, expr2.type))

        if ctx.STEP():
            expr3 = self.visit(ctx.expr(2))
            if expr3.type != Type.int():
                raise CompileError("Expression {} should have type int, but is {}".format(expr3, expr3.type))
            step = expr3.value
        else:
            step = 1

        for i in range(expr1.value, expr2.value + 1, step):
            self._get_latest_symbol_table().update(ident, i)
            # scope - keep updates on existing variables, discard new variables
            self._visit_commands_with_scope(ctx.commands())

    def visitRepeat(self, ctx: xDroneParser.RepeatContext) -> None:
        expr = self.visit(ctx.expr())
        if expr.type != Type.int():
            raise CompileError("Expression {} should have type int, but is {}".format(expr, expr.type))
        times = expr.value

        for _ in range(times):
            # scope - keep updates on existing variables, discard new variables
            self._visit_commands_with_scope(ctx.commands())

    def visitReturn(self, ctx: xDroneParser.ReturnContext) -> None:
        if len(self.returned) == 1:
            raise CompileError("Cannot return in the Main function")
        self.returned[-1] = True  # order important
        if ctx.expr():
            expr = self.visit(ctx.expr())
            self.returned_value[-1] = Expression(expr.type, expr.value, ident=None)

    def visitParallel(self, ctx: xDroneParser.ParallelContext) -> None:
        parallel_commands = ParallelDroneCommands()
        for commands in ctx.commands():
            self.commands.append([])

            # scope - discard updates on existing variables, discard new variables
            new_symbol_table = copy.deepcopy(self._get_latest_symbol_table())
            self.symbol_table.append(new_symbol_table)
            self.returned.append(False)
            self.returned_value.append(None)
            self.visit(commands)
            returned_value = self.returned_value.pop(-1)
            self.returned.pop(-1)
            self.symbol_table.pop(-1)
            if returned_value is not None:
                raise CompileError("Parallel branch should not return anything, but {} is returned"
                                   .format(returned_value))

            branch = self.commands.pop(-1)
            try:
                parallel_commands.add(branch)
            except RepeatDroneNameException as e:
                raise CompileError("Parallel branches should have exclusive drone names, "
                                   "but {} appeared in more than one branches"
                                   .format(e.repeated_names))

        self._get_latest_commands().append(parallel_commands)

    ######## ident ########

    def visitIdent(self, ctx: xDroneParser.IdentContext) -> Identifier:
        ident = ctx.IDENT().getText()
        # ident in symbol table will shadow drone constant
        if ident in self._get_latest_symbol_table():
            return Identifier(str(ident), self._get_latest_symbol_table().get_expression(ident))
        if ident in self.drones:
            return Identifier(str(ident), Expression(Type.drone(), self.drones[ident]))
        return Identifier(str(ident), None)

    ######## list_elem ########

    def visitListElem(self, ctx: xDroneParser.ListElemContext) -> ListElem:
        expr1, expr2 = self.visit(ctx.expr(0)), self.visit(ctx.expr(1))
        if not isinstance(expr1.type, ListType):
            raise CompileError("Expression {} should have type list, but is {}".format(expr1, expr1.type))
        if expr2.type != Type.int():
            raise CompileError("Expression {} should have type int, but is {}".format(expr2, expr2.type))
        if expr2.value >= len(expr1.value) or expr2.value < 0:
            raise CompileError("List {} has length {}, but has been assessed with out-of-range index {}"
                               .format(expr1, len(expr1.value), expr2.value))
        return ListElem(expr1.ident, expr1, expr2.value)

    ######## vector_elem ########

    def visitVectorX(self, ctx: xDroneParser.VectorXContext) -> VectorElem:
        expr = self.visit(ctx.expr())
        if expr.type != Type.vector():
            raise CompileError("Expression {} should have type vector, but is {}".format(expr, expr.type))
        return VectorElem(expr.ident, expr, 0)

    def visitVectorY(self, ctx: xDroneParser.VectorYContext) -> VectorElem:
        expr = self.visit(ctx.expr())
        if expr.type != Type.vector():
            raise CompileError("Expression {} should have type vector, but is {}".format(expr, expr.type))
        return VectorElem(expr.ident, expr, 1)

    def visitVectorZ(self, ctx: xDroneParser.VectorZContext) -> VectorElem:
        expr = self.visit(ctx.expr())
        if expr.type != Type.vector():
            raise CompileError("Expression {} should have type vector, but is {}".format(expr, expr.type))
        return VectorElem(expr.ident, expr, 2)

    ######## functions ########

    def visitFuncIdent(self, ctx: xDroneParser.FuncIdentContext):
        ident = ctx.IDENT().getText()
        return FunctionIdentifier(str(ident))

    def visitCall(self, ctx: xDroneParser.CallContext) -> Optional[Expression]:
        func_identifier = self.visit(ctx.funcIdent())
        ident = func_identifier.ident
        arg_list = self.visit(ctx.argList()) if ctx.argList() else []
        if ident not in self.function_table:
            raise CompileError("Function or procedure {} has not been defined".format(ident))
        function = self.function_table.get_function(ident)
        arg_types = [arg.type for arg in arg_list]
        param_types = [param.type for param in function.param_list]
        if arg_types != param_types:
            raise CompileError("Arguments when calling function or procedure {} should have types {}, but is {}"
                               .format(ident, [str(type) for type in param_types], [str(type) for type in arg_types]))
        new_symbol_table = SymbolTable()
        param_idents = [param.ident for param in function.param_list]
        for param_ident, expr in zip(param_idents, arg_list):
            new_symbol_table.store(param_ident, expr)

        self.symbol_table.append(new_symbol_table)
        self.returned.append(False)
        self.returned_value.append(None)
        self.visit(function.get_commands())
        returned_value = self.returned_value.pop(-1)
        self.returned.pop(-1)
        self.symbol_table.pop(-1)
        if function.return_type is None:
            if returned_value is not None:
                raise CompileError("Procedure {} should not return anything, but {} is returned"
                                   .format(ident, returned_value))
            return None
        else:
            if returned_value is None:
                raise CompileError("Function {} has returned type {}, but nothing is returned"
                                   .format(ident, function.return_type))
            if returned_value.type != function.return_type:
                raise CompileError("Function {} has returned type {}, but {} is returned"
                                   .format(ident, function.return_type, returned_value.type))
            return returned_value

    def visitArgList(self, ctx: xDroneParser.ArgListContext) -> List[Expression]:
        exprs = [self.visit(expr) for expr in ctx.expr()]
        return exprs

    def visitFunction(self, ctx: xDroneParser.FunctionContext) -> None:
        func_identifier, return_type = self.visit(ctx.funcIdent()), self.visit(ctx.type_())
        ident = func_identifier.ident
        if ident in self.function_table:
            raise CompileError("Function or procedure {} already defined".format(ident))
        param_list = self.visit(ctx.paramList()) if ctx.paramList() else []
        self.function_table.store(ident, Function(ident, param_list, return_type, ctx.commands()))

    def visitProcedure(self, ctx: xDroneParser.ProcedureContext) -> None:
        func_identifier = self.visit(ctx.funcIdent())
        ident = func_identifier.ident
        if ident in self.function_table:
            raise CompileError("Function or procedure {} already defined".format(ident))
        param_list = self.visit(ctx.paramList()) if ctx.paramList() else []
        self.function_table.store(ident, Function(ident, param_list, None, ctx.commands()))

    def visitParamList(self, ctx: xDroneParser.ParamListContext) -> List[Parameter]:
        types = [self.visit(type) for type in ctx.type_()]
        idents = [self.visit(ident).ident for ident in ctx.ident()]
        if len(idents) != len(set(idents)):
            raise CompileError("Parameter names are duplicated in {}".format(idents))
        parameters = []
        for type, ident in zip(types, idents):
            parameters.append(Parameter(ident, type))
        return parameters

    ######## type ########

    def visitIntType(self, ctx: xDroneParser.IntTypeContext) -> Type:
        return Type.int()

    def visitDecimalType(self, ctx: xDroneParser.DecimalTypeContext) -> Type:
        return Type.decimal()

    def visitStringType(self, ctx: xDroneParser.StringTypeContext) -> Type:
        return Type.string()

    def visitBooleanType(self, ctx: xDroneParser.BooleanTypeContext) -> Type:
        return Type.boolean()

    def visitVectorType(self, ctx: xDroneParser.VectorTypeContext) -> Type:
        return Type.vector()

    def visitDroneType(self, ctx: xDroneParser.DroneTypeContext) -> Type:
        return Type.drone()

    def visitListType(self, ctx: xDroneParser.ListTypeContext) -> Type:
        elem_type = self.visit(ctx.type_())
        return Type.list_of(elem_type)

    ######## expr ########

    def visitIntExpr(self, ctx: xDroneParser.IntExprContext) -> Expression:
        signed_int = ctx.INT().getText()
        return Expression(Type.int(), int(signed_int))

    def visitDecimalExpr(self, ctx: xDroneParser.DecimalExprContext) -> Expression:
        signed_float = ctx.FLOAT().getText()
        return Expression(Type.decimal(), float(signed_float))

    def visitStringExpr(self, ctx: xDroneParser.StringExprContext) -> Expression:
        escaped_string = ctx.ESCAPED_STRING().getText()
        quotation_removed = str(escaped_string)[1:-1]
        unescaped = bytes(quotation_removed, "utf-8").decode("unicode_escape")
        return Expression(Type.string(), unescaped)

    def visitTrueExpr(self, ctx: xDroneParser.TrueExprContext) -> Expression:
        return Expression(Type.boolean(), True)

    def visitFalseExpr(self, ctx: xDroneParser.FalseExprContext) -> Expression:
        return Expression(Type.boolean(), False)

    def visitIdentExpr(self, ctx: xDroneParser.IdentExprContext) -> Expression:
        identifier = self.visit(ctx.ident())
        ident = identifier.ident
        if ident not in self._get_latest_symbol_table() and ident not in self.drones:
            # identifier.expression is None iff child.ident not in latest symbol table nor in drones constants
            raise CompileError("Identifier {} has not been declared".format(ident))
        return identifier.to_expression()

    def visitListElemExpr(self, ctx: xDroneParser.ListElemExprContext) -> Expression:
        expr1, expr2 = self.visit(ctx.expr(0)), self.visit(ctx.expr(1))
        if not isinstance(expr1.type, ListType):
            raise CompileError("Expression {} should have type list, but is {}".format(expr1, expr1.type))
        if expr2.type != Type.int():
            raise CompileError("Expression {} should have type int, but is {}".format(expr2, expr2.type))
        if expr2.value >= len(expr1.value) or expr2.value < 0:
            raise CompileError("List {} has length {}, but has been assessed with out-of-range index {}"
                               .format(expr1, len(expr1.value), expr2.value))
        return ListElem(expr1.ident, expr1, expr2.value).to_expression()

    def visitVectorXExpr(self, ctx: xDroneParser.VectorXExprContext) -> Expression:
        expr = self.visit(ctx.expr())
        if expr.type != Type.vector():
            raise CompileError("Expression {} should have type vector, but is {}".format(expr, expr.type))
        return VectorElem(expr.ident, expr, 0).to_expression()

    def visitVectorYExpr(self, ctx: xDroneParser.VectorYExprContext) -> Expression:
        expr = self.visit(ctx.expr())
        if expr.type != Type.vector():
            raise CompileError("Expression {} should have type vector, but is {}".format(expr, expr.type))
        return VectorElem(expr.ident, expr, 1).to_expression()

    def visitVectorZExpr(self, ctx: xDroneParser.VectorZExprContext) -> Expression:
        expr = self.visit(ctx.expr())
        if expr.type != Type.vector():
            raise CompileError("Expression {} should have type vector, but is {}".format(expr, expr.type))
        return VectorElem(expr.ident, expr, 2).to_expression()

    def visitList(self, ctx: xDroneParser.ListContext) -> Expression:
        exprs = [self.visit(expr) for expr in ctx.expr()]
        if len(exprs) == 0:
            return Expression(Type.empty_list(), [])
        if not all(e.type == exprs[0].type for e in exprs):
            raise CompileError("Elements in list {} should have the same type".format([str(e) for e in exprs]))
        return Expression(Type.list_of(exprs[0].type), [e.value for e in exprs])

    def visitVector(self, ctx: xDroneParser.VectorContext) -> Expression:
        expr1, expr2, expr3 = self.visit(ctx.expr(0)), self.visit(ctx.expr(1)), self.visit(ctx.expr(2))
        for expr in [expr1, expr2, expr3]:
            if expr.type != Type.int() and expr.type != Type.decimal():
                raise CompileError("Expression {} should have type int or decimal, but is {}".format(expr, expr.type))
        return Expression(Type.vector(), [float(expr1.value), float(expr2.value), float(expr3.value)])

    def visitFunctionCall(self, ctx: xDroneParser.FunctionCallContext) -> Expression:
        call = self.visit(ctx.call())
        if call is None:
            raise CompileError("Function call should return an expression, but nothing is returned")
        return call

    def visitSize(self, ctx: xDroneParser.SizeContext) -> Expression:
        expr = self.visit(ctx.expr())
        if not isinstance(expr.type, ListType):
            raise CompileError("Expression {} should have type list, but is {}".format(expr, expr.type))
        return Expression(Type.int(), len(expr.value))

    def visitParentheses(self, ctx: xDroneParser.ParenthesesContext) -> Expression:
        expr = self.visit(ctx.expr())
        return expr

    def visitPositNegate(self, ctx: xDroneParser.PositNegateContext) -> Expression:
        expr = self.visit(ctx.expr())
        if expr.type == Type.int() or expr.type == Type.decimal():
            if ctx.PLUS():
                result_value = +expr.value
            else:  # MINUS
                result_value = -expr.value
        elif expr.type == Type.vector():
            if ctx.PLUS():
                result_value = [+e for e in expr.value]
            else:  # MINUS
                result_value = [-e for e in expr.value]
        else:
            raise CompileError(
                "Expression {} should have type int, decimal or vector, but is {}".format(expr, expr.type))
        return Expression(expr.type, result_value)

    def visitNot(self, ctx: xDroneParser.NotContext) -> Expression:
        expr = self.visit(ctx.expr())
        if expr.type != Type.boolean():
            raise CompileError("Expression {} should have type boolean, but is {}".format(expr, expr.type))
        return Expression(Type.boolean(), not expr.value)

    def _is_int_or_decimal(self, type: Type):
        return type == Type.int() or type == Type.decimal()

    def _get_int_decimal_result_type(self, type1: Type, type2: Type):
        assert self._is_int_or_decimal(type1)
        assert self._is_int_or_decimal(type2)
        if type1 == Type.decimal() or type2 == Type.decimal():
            return Type.decimal(), float
        return Type.int(), int

    def visitMultiDivide(self, ctx: xDroneParser.MultiDivideContext) -> Expression:
        expr1, expr2 = self.visit(ctx.expr(0)), self.visit(ctx.expr(1))
        if self._is_int_or_decimal(expr1.type) and self._is_int_or_decimal(expr2.type):
            result_type, func = self._get_int_decimal_result_type(expr1.type, expr2.type)
            if ctx.MULTI():
                result_value = func(expr1.value * expr2.value)
            else:  # DIV
                if expr2.value == 0:
                    raise CompileError("Division by zero")
                result_value = func(expr1.value / expr2.value)
        elif self._is_int_or_decimal(expr1.type) and expr2.type == Type.vector():
            result_type = Type.vector()
            if ctx.MULTI():
                result_value = [expr1.value * e for e in expr2.value]
            else:  # DIV
                raise CompileError("Expression {} and {} have wrong types to perform multiplication or division"
                                   .format(expr1, expr2))
        elif expr1.type == Type.vector() and self._is_int_or_decimal(expr2.type):
            result_type = Type.vector()
            if ctx.MULTI():
                result_value = [e * expr2.value for e in expr1.value]
            else:  # DIV
                result_value = [e / expr2.value for e in expr1.value]
        else:
            raise CompileError("Expression {} and {} have wrong types to perform multiplication or division"
                               .format(expr1, expr2))
        return Expression(result_type, result_value)

    def visitPlusMinus(self, ctx: xDroneParser.PlusMinusContext) -> Expression:
        expr1, expr2 = self.visit(ctx.expr(0)), self.visit(ctx.expr(1))
        if self._is_int_or_decimal(expr1.type) and self._is_int_or_decimal(expr2.type):
            result_type, func = self._get_int_decimal_result_type(expr1.type, expr2.type)
            if ctx.PLUS():
                result_value = func(expr1.value + expr2.value)
            else:  # MINUS
                result_value = func(expr1.value - expr2.value)
        elif expr1.type == Type.vector() and expr2.type == Type.vector():
            result_type = Type.vector()
            if ctx.PLUS():
                result_value = [e1 + e2 for e1, e2 in zip(expr1.value, expr2.value)]
            else:  # MINUS
                result_value = [e1 - e2 for e1, e2 in zip(expr1.value, expr2.value)]
        else:
            raise CompileError("Expression {} and {} have wrong types to perform addition or subtraction"
                               .format(expr1, expr2))
        return Expression(result_type, result_value)

    def visitConcat(self, ctx: xDroneParser.ConcatContext) -> Expression:
        expr1, expr2 = self.visit(ctx.expr(0)), self.visit(ctx.expr(1))
        if expr1.type != Type.string():
            raise CompileError("Expression {} should have type string, but is {}".format(expr1, expr1.type))
        if expr2.type != Type.string():
            raise CompileError("Expression {} should have type string, but is {}".format(expr2, expr2.type))
        return Expression(Type.string(), expr1.value + expr2.value)

    def visitCompare(self, ctx: xDroneParser.CompareContext) -> Expression:
        expr1, expr2 = self.visit(ctx.expr(0)), self.visit(ctx.expr(1))
        if expr1.type != Type.int() and expr1.type != Type.decimal():
            raise CompileError("Expression {} should have type int or decimal, but is {}".format(expr1, expr1.type))
        if expr2.type != Type.int() and expr2.type != Type.decimal():
            raise CompileError("Expression {} should have type int or decimal, but is {}".format(expr2, expr2.type))
        if ctx.GREATER():
            result_value = expr1.value > expr2.value
        elif ctx.GREATER_EQ():
            result_value = expr1.value >= expr2.value
        elif ctx.LESS():
            result_value = expr1.value < expr2.value
        else:  # LESS_EQ
            result_value = expr1.value <= expr2.value
        return Expression(Type.boolean(), result_value)

    def visitEquality(self, ctx: xDroneParser.EqualityContext) -> Expression:
        expr1, expr2 = self.visit(ctx.expr(0)), self.visit(ctx.expr(1))
        if expr1.type != expr2.type:
            raise CompileError("Expressions {} and {} should have the same type".format(expr1, expr2))
        if ctx.EQ():
            result_value = expr1.value == expr2.value
        else:  # NOT_EQ
            result_value = expr1.value != expr2.value
        return Expression(Type.boolean(), result_value)

    def visitAnd(self, ctx: xDroneParser.AndContext) -> Expression:
        expr1, expr2 = self.visit(ctx.expr(0)), self.visit(ctx.expr(1))
        if expr1.type != Type.boolean():
            raise CompileError("Expression {} should have type boolean, but is {}".format(expr1, expr1.type))
        if expr2.type != Type.boolean():
            raise CompileError("Expression {} should have type boolean, but is {}".format(expr2, expr2.type))
        return Expression(Type.boolean(), expr1.value and expr2.value)

    def visitOr(self, ctx: xDroneParser.OrContext) -> Expression:
        expr1, expr2 = self.visit(ctx.expr(0)), self.visit(ctx.expr(1))
        if expr1.type != Type.boolean():
            raise CompileError("Expression {} should have type boolean, but is {}".format(expr1, expr1.type))
        if expr2.type != Type.boolean():
            raise CompileError("Expression {} should have type boolean, but is {}".format(expr2, expr2.type))
        return Expression(Type.boolean(), expr1.value or expr2.value)



from typing import Dict

import antlr4

from antlr.xDroneLexer import xDroneLexer, CommonTokenStream
from antlr.xDroneParser import xDroneParser
from xdrone.compiler.compiler import Compiler
from xdrone.compiler.compiler_utils.drones import Drone
from xdrone.compiler.compiler_utils.error_listener import ParserErrorListener
from xdrone.compiler.compiler_utils.expressions import Expression
from xdrone.compiler.compiler_utils.functions import FunctionTable
from xdrone.compiler.compiler_utils.symbol_table import SymbolTable
from xdrone.config_parsers.config_parser import ConfigParser
from xdrone.safety_checker.boundary_checker import BoundaryChecker
from xdrone.safety_checker.collision_checker import CollisionChecker
from xdrone.shared.boundary_config import BoundaryConfig
from xdrone.shared.collision_config import DefaultCollisionConfig
from xdrone.shared.compile_error import XDroneSyntaxError
from xdrone.shared.drone_config import DefaultDroneConfig, DroneConfig
from xdrone.shared.safety_check_error import SafetyCheckError
from xdrone.state_updaters.state_updater import StateUpdater


def generate_commands(program, drone_config_map: Dict[str, DroneConfig] = None,
                      boundary_checker: BoundaryChecker = None, collision_checker: CollisionChecker = None,
                      has_checks: bool = True, save_report: bool = False,
                      symbol_table: SymbolTable = None, function_table: FunctionTable = None):
    if drone_config_map is None:
        drone_config_map = {"DEFAULT": DefaultDroneConfig()}
    if boundary_checker is None:
        boundary_checker = BoundaryChecker(BoundaryConfig.no_limit())
    if collision_checker is None:
        collision_checker = CollisionChecker(drone_config_map, DefaultCollisionConfig())
    if symbol_table is None:
        symbol_table = SymbolTable()
    if function_table is None:
        function_table = FunctionTable()
    state_updater_map = {name: StateUpdater(config) for name, config in drone_config_map.items()}

    tree = _parse_program(program)

    drone_commands = Compiler(drone_config_map, symbol_table, function_table).visit(tree)

    if has_checks:
        boundary_checker.check(drone_commands, state_updater_map)
        collision_checker.check(drone_commands, state_updater_map, save_report)

    return drone_commands


def _parse_program(program):
    input_stream = antlr4.InputStream(program)
    # lexing
    lexer = xDroneLexer(input_stream)
    stream = CommonTokenStream(lexer)
    # parsing
    parser = xDroneParser(stream)
    parser.removeErrorListeners()
    error_listener = ParserErrorListener()
    parser.addErrorListener(error_listener)
    tree = parser.prog()
    if error_listener.syntax_errors:
        raise XDroneSyntaxError(error_listener.get_error_string())
    return tree


def generate_commands_with_config(program, config_json, has_checks, save_report):
    drone_config_map, boundary_config, collision_config = ConfigParser.parse(config_json)
    boundary_checker = BoundaryChecker(boundary_config)
    collision_checker = CollisionChecker(drone_config_map, collision_config)
    generated_commands = generate_commands(program, drone_config_map, boundary_checker, collision_checker,
                                           has_checks, save_report)
    return generated_commands, drone_config_map, boundary_config



import unittest

from xdrone.command_generators import generate_commands
from xdrone.compiler.compiler_utils.expressions import Expression
from xdrone.compiler.compiler_utils.type import Type
from xdrone.shared.command import Command, SingleDroneCommand, ParallelDroneCommands
from xdrone.shared.compile_error import CompileError
from xdrone.shared.drone_config import DefaultDroneConfig


class ParallelTest(unittest.TestCase):
    def test_parallel_with_different_drones_should_give_correct_commands(self):
        drone_config_map = {"DRONE1": DefaultDroneConfig(),
                            "DRONE2": DefaultDroneConfig(),
                            "DRONE3": DefaultDroneConfig()}
        actual = generate_commands("""
            main() { 
              {DRONE1.takeoff();} || {DRONE2.takeoff();} || {DRONE3.takeoff();}; 
              {
                DRONE1.land();
              } || {
                {DRONE2.land();} || {DRONE3.land();};
              };
            }
        """, drone_config_map=drone_config_map)
        expected = [ParallelDroneCommands([[SingleDroneCommand("DRONE1", Command.takeoff())],
                                           [SingleDroneCommand("DRONE2", Command.takeoff())],
                                           [SingleDroneCommand("DRONE3", Command.takeoff())]]),
                    ParallelDroneCommands([[SingleDroneCommand("DRONE1", Command.land())],
                                           [ParallelDroneCommands([
                                               [SingleDroneCommand("DRONE2", Command.land())],
                                               [SingleDroneCommand("DRONE3", Command.land())]
                                           ])]])]
        self.assertEqual(expected, actual)

    def test_parallel_with_repeated_drones_in_branches_should_give_error(self):
        drone_config_map = {"DRONE1": DefaultDroneConfig(),
                            "DRONE2": DefaultDroneConfig(),
                            "DRONE3": DefaultDroneConfig()}
        with self.assertRaises(CompileError) as context:
            generate_commands("""
                procedure foo() {
                  DRONE1.takeoff();
                  DRONE2.takeoff();
                  DRONE1.land();
                  DRONE2.land();
                }
                procedure bar() {
                  DRONE2.takeoff();
                  DRONE3.takeoff();
                  DRONE2.land();
                  DRONE3.land();
                }
                main() { 
                  {foo();} || {bar();} || {DRONE3.takeoff(); DRONE3.land();}; 
                }
            """, drone_config_map=drone_config_map)
        self.assertTrue("Parallel branches should have exclusive drone names, " +
                        "but {'DRONE2'} appeared in more than one branches"
                        in str(context.exception))

    def test_parallel_return_in_branch_should_return_early(self):
        drone_config_map = {"DRONE1": DefaultDroneConfig(),
                            "DRONE2": DefaultDroneConfig(),
                            "DRONE3": DefaultDroneConfig()}
        actual = generate_commands("""
            main() { 
              {return; DRONE1.takeoff();} || {DRONE2.takeoff(); DRONE2.land();}; 
            }
        """, drone_config_map=drone_config_map)
        expected = [ParallelDroneCommands([[],
                                           [SingleDroneCommand("DRONE2", Command.takeoff()),
                                            SingleDroneCommand("DRONE2", Command.land())]])]
        self.assertEqual(expected, actual)

    def test_parallel_return_with_value_in_branch_should_give_error(self):
        drone_config_map = {"DRONE1": DefaultDroneConfig(),
                            "DRONE2": DefaultDroneConfig(),
                            "DRONE3": DefaultDroneConfig()}
        with self.assertRaises(CompileError) as context:
            generate_commands("""
                main() { 
                  {return 1; DRONE1.takeoff();} || {DRONE2.takeoff(); DRONE2.land();}; 
                }
            """, drone_config_map=drone_config_map)
        self.assertTrue("Parallel branch should not return anything, but {} is returned"
                        .format(Expression(Type.int(), 1))
                        in str(context.exception))



import unittest
from unittest.mock import patch, mock_open, Mock

from xdrone.safety_checker.collision_checker import CollisionChecker
from xdrone.shared.collision_config import CollisionConfig
from xdrone.shared.command import Command, SingleDroneCommand, ParallelDroneCommands
from xdrone.shared.drone_config import DroneConfig
from xdrone.shared.safety_check_error import SafetyCheckError
from xdrone.state_updaters.state_updater import StateUpdater


class CollisionCheckerTest(unittest.TestCase):
    def setUp(self) -> None:
        self.collision_config = CollisionConfig(collision_meters=0.3, time_interval_seconds=0.1)
        self.drone_config_map = {"DRONE1": DroneConfig(init_position=(0, 0, 0), speed_mps=1,
                                                       rotate_speed_dps=90, takeoff_height_meters=1),
                                 "DRONE2": DroneConfig(init_position=(1, 0, 0), speed_mps=2,
                                                       rotate_speed_dps=180, takeoff_height_meters=1)}
        self.state_updater_map = {"DRONE1": StateUpdater(DroneConfig(init_position=(0, 0, 0), speed_mps=1,
                                                                     rotate_speed_dps=90, takeoff_height_meters=1)),
                                  "DRONE2": StateUpdater(DroneConfig(init_position=(1, 0, 0), speed_mps=2,
                                                                     rotate_speed_dps=180, takeoff_height_meters=1))}

        self.collision_checker = CollisionChecker(self.drone_config_map, self.collision_config)

    def test_check_no_collision_should_not_give_error(self):
        drone_commands = [SingleDroneCommand("DRONE1", Command.takeoff()),
                          SingleDroneCommand("DRONE1", Command.land())]
        self.collision_checker.check(drone_commands, self.state_updater_map)

    def test_check_bad_collision_config_should_not_give_error(self):
        collision_config = CollisionConfig(collision_meters=0.3, time_interval_seconds=0.001)
        collision_checker = CollisionChecker(self.drone_config_map, collision_config)
        drone_commands = [SingleDroneCommand("DRONE1", Command.takeoff()),
                          SingleDroneCommand("DRONE1", Command.land())]
        with self.assertRaises(SafetyCheckError) as context:
            collision_checker.check(drone_commands, self.state_updater_map)
        self.assertTrue("Error occurred during collision check, please retry with a better collision_config."
                        in str(context.exception))

    def test_check_with_large_variance_should_perform_not_like_determined(self):
        collision_config = CollisionConfig(collision_meters=0.3, time_interval_seconds=0.1, confidence_threshold=0.95)
        drone_config_map = {"DRONE1": DroneConfig(init_position=(0, 0, 0), speed_mps=1,
                                                  rotate_speed_dps=90, takeoff_height_meters=1,
                                                  var_per_meter=100),
                            "DRONE2": DroneConfig(init_position=(1, 0, 0), speed_mps=2,
                                                  rotate_speed_dps=180, takeoff_height_meters=1,
                                                  var_per_meter=100)}
        collision_checker = CollisionChecker(drone_config_map, collision_config)

        drone_commands = [SingleDroneCommand("DRONE1", Command.takeoff()),
                          SingleDroneCommand("DRONE2", Command.takeoff()),
                          SingleDroneCommand("DRONE1", Command.right(1)),
                          SingleDroneCommand("DRONE1", Command.left(1)),
                          SingleDroneCommand("DRONE1", Command.land()),
                          SingleDroneCommand("DRONE2", Command.land())]

        collision_checker.check(drone_commands, self.state_updater_map)

    def test_check_with_small_variance_should_perform_similar_to_determined(self):
        collision_config = CollisionConfig(collision_meters=0.3, time_interval_seconds=0.1, confidence_threshold=0.99)
        drone_config_map = {"DRONE1": DroneConfig(init_position=(0, 0, 0), speed_mps=1,
                                                  rotate_speed_dps=90, takeoff_height_meters=1,
                                                  var_per_meter=0.001),
                            "DRONE2": DroneConfig(init_position=(1, 0, 0), speed_mps=2,
                                                  rotate_speed_dps=180, takeoff_height_meters=1,
                                                  var_per_meter=0.001)}
        collision_checker = CollisionChecker(drone_config_map, collision_config)

        drone_commands = [SingleDroneCommand("DRONE1", Command.takeoff()),
                          SingleDroneCommand("DRONE2", Command.takeoff()),
                          SingleDroneCommand("DRONE1", Command.right(1)),
                          SingleDroneCommand("DRONE1", Command.left(1)),
                          SingleDroneCommand("DRONE1", Command.land()),
                          SingleDroneCommand("DRONE2", Command.land())]

        with self.assertRaises(SafetyCheckError) as context:
            collision_checker.check(drone_commands, self.state_updater_map)
        self.assertTrue("Collision might happen between DRONE1 and DRONE2, at time 2.4s, " +
                        "near position (x=0.95m, y=0.0m, z=1.0m), distance=0.1m, confidence=99.968%"
                        in str(context.exception))
        self.assertTrue("Collision might happen between DRONE1 and DRONE2, at time 2.6s, " +
                        "near position (x=0.95m, y=0.0m, z=1.0m), distance=0.1m, confidence=99.949%"
                        in str(context.exception))

    def test_check_with_zero_variance_should_perform_as_determined(self):
        collision_config = CollisionConfig(collision_meters=0.3, time_interval_seconds=0.1, confidence_threshold=1.0)
        drone_config_map = {"DRONE1": DroneConfig(init_position=(0, 0, 0), speed_mps=1,
                                                  rotate_speed_dps=90, takeoff_height_meters=1,
                                                  var_per_meter=0),
                            "DRONE2": DroneConfig(init_position=(1, 0, 0), speed_mps=2,
                                                  rotate_speed_dps=180, takeoff_height_meters=1,
                                                  var_per_meter=0)}
        collision_checker = CollisionChecker(drone_config_map, collision_config)

        drone_commands = [SingleDroneCommand("DRONE1", Command.takeoff()),
                          SingleDroneCommand("DRONE2", Command.takeoff()),
                          SingleDroneCommand("DRONE1", Command.right(1)),
                          SingleDroneCommand("DRONE1", Command.left(1)),
                          SingleDroneCommand("DRONE1", Command.land()),
                          SingleDroneCommand("DRONE2", Command.land())]

        with self.assertRaises(SafetyCheckError) as context:
            collision_checker.check(drone_commands, self.state_updater_map)
        self.assertTrue("Collision might happen between DRONE1 and DRONE2, at time 2.2s, " +
                        "near position (x=0.85m, y=0.0m, z=1.0m), distance=0.3m, confidence=100.000%"
                        in str(context.exception))
        self.assertTrue("Collision might happen between DRONE1 and DRONE2, at time 2.8s, " +
                        "near position (x=0.85m, y=0.0m, z=1.0m), distance=0.3m, confidence=100.000%"
                        in str(context.exception))

    def test_check_single_drone_command_should_detect_collision_and_give_error(self):
        drone_commands = [SingleDroneCommand("DRONE1", Command.takeoff()),
                          SingleDroneCommand("DRONE2", Command.takeoff()),
                          SingleDroneCommand("DRONE1", Command.right(1)),
                          SingleDroneCommand("DRONE2", Command.wait(1)),
                          SingleDroneCommand("DRONE2", Command.rotate_left(90)),
                          SingleDroneCommand("DRONE2", Command.forward(1)),
                          SingleDroneCommand("DRONE1", Command.land()),
                          SingleDroneCommand("DRONE2", Command.land())]

        with self.assertRaises(SafetyCheckError) as context:
            self.collision_checker.check(drone_commands, self.state_updater_map)
        self.assertTrue("Collisions might happen!\nCollision might happen between DRONE1 and DRONE2"
                        in str(context.exception))

    def test_check_parallel_drone_command_should_detect_collision_and_give_error(self):
        drone_commands = [SingleDroneCommand("DRONE1", Command.takeoff()),
                          SingleDroneCommand("DRONE2", Command.takeoff()),
                          ParallelDroneCommands([
                              [SingleDroneCommand("DRONE1", Command.right(1))],
                              [SingleDroneCommand("DRONE2", Command.left(1))]
                          ]),
                          SingleDroneCommand("DRONE1", Command.land()),
                          SingleDroneCommand("DRONE2", Command.land())]

        with self.assertRaises(SafetyCheckError) as context:
            self.collision_checker.check(drone_commands, self.state_updater_map)
        self.assertTrue("Collisions might happen!\nCollision might happen between DRONE1 and DRONE2"
                        in str(context.exception))

    @patch('xdrone.safety_checker.collision_checker.CollisionReportSaver.save_check_report')
    def test_check_with_save_log_flag_should_call_log_saver(self, mock_save_report):
        drone_commands = [SingleDroneCommand("DRONE1", Command.takeoff()),
                          SingleDroneCommand("DRONE1", Command.land())]
        self.collision_checker.check(drone_commands, self.state_updater_map, save_report=True)
        mock_save_report.assert_called()


class CollisionReportSaverTest(unittest.TestCase):
    def setUp(self) -> None:
        self.collision_config = CollisionConfig(collision_meters=0.3, time_interval_seconds=0.1)
        self.drone_config_map = {"DRONE1": DroneConfig(init_position=(0, 0, 0), speed_mps=1,
                                                       rotate_speed_dps=90, takeoff_height_meters=1),
                                 "DRONE2": DroneConfig(init_position=(1, 0, 0), speed_mps=2,
                                                       rotate_speed_dps=180, takeoff_height_meters=1)}
        self.state_updater_map = {"DRONE1": StateUpdater(DroneConfig(init_position=(0, 0, 0), speed_mps=1,
                                                                     rotate_speed_dps=90, takeoff_height_meters=1)),
                                  "DRONE2": StateUpdater(DroneConfig(init_position=(1, 0, 0), speed_mps=2,
                                                                     rotate_speed_dps=180, takeoff_height_meters=1))}

        self.collision_checker = CollisionChecker(self.drone_config_map, self.collision_config)

    @patch('os.path.isdir')
    @patch('os.path.join')
    @patch('os.mkdir')
    @patch('builtins.open', new_callable=mock_open())
    @patch('matplotlib.pyplot.subplots')
    def test_save_check_report_should_save_report(self, mock_subplots, mock_open_file, mock_mkdir, mock_join, mock_isdir):
        drone_commands = [SingleDroneCommand("DRONE1", Command.takeoff()),
                          SingleDroneCommand("DRONE1", Command.land())]
        mock_isdir.return_value = False
        mock_join.return_value = "joined_path"
        mock_subplots.return_value = (Mock(), Mock())

        self.collision_checker.check(drone_commands, self.state_updater_map, save_report=True)

        mock_mkdir.assert_called_with("reports")
        mock_open_file.assert_called_with("joined_path", "w")



import unittest

from xdrone.command_generators import generate_commands
from xdrone.compiler.compiler_utils.expressions import Expression
from xdrone.compiler.compiler_utils.type import Type
from xdrone.shared.command import Command, SingleDroneCommand
from xdrone.shared.compile_error import CompileError
from xdrone.shared.drone_config import DefaultDroneConfig


class MovementCommandsTest(unittest.TestCase):
    def test_takeoff_should_return_correct_command(self):
        actual = generate_commands("""
            main() { takeoff(); land(); }
        """)
        expected = [SingleDroneCommand("DEFAULT", Command.takeoff()),
                    SingleDroneCommand("DEFAULT", Command.land())]
        self.assertEqual(expected, actual)

    def test_takeoff_with_drone_name_should_return_correct_command(self):
        actual = generate_commands("""
            main() { DRONE1.takeoff(); DRONE1.land(); }
        """, drone_config_map={"DRONE1": DefaultDroneConfig()})
        expected = [SingleDroneCommand("DRONE1", Command.takeoff()),
                    SingleDroneCommand("DRONE1", Command.land())]
        self.assertEqual(expected, actual)

    def test_land_should_return_correct_command(self):
        actual = generate_commands("""
            main() { takeoff(); land(); }
        """)
        expected = [SingleDroneCommand("DEFAULT", Command.takeoff()),
                    SingleDroneCommand("DEFAULT", Command.land())]
        self.assertEqual(expected, actual)

    def test_land_with_drone_name_should_return_correct_command(self):
        actual = generate_commands("""
            main() { DRONE1.takeoff(); DRONE1.land(); }
        """, drone_config_map={"DRONE1": DefaultDroneConfig()})
        expected = [SingleDroneCommand("DRONE1", Command.takeoff()),
                    SingleDroneCommand("DRONE1", Command.land())]
        self.assertEqual(expected, actual)

    def test_up_with_int_parameter_should_return_correct_command(self):
        actual = generate_commands("""
            main() { takeoff(); up(1); land(); }
        """)
        expected = [SingleDroneCommand("DEFAULT", Command.takeoff()),
                    SingleDroneCommand("DEFAULT", Command.up(1)),
                    SingleDroneCommand("DEFAULT", Command.land())]
        self.assertEqual(expected, actual)

    def test_up_with_decimal_parameter_should_return_correct_command(self):
        actual = generate_commands("""
            main() { takeoff(); up(1.0); land(); }
        """)
        expected = [SingleDroneCommand("DEFAULT", Command.takeoff()),
                    SingleDroneCommand("DEFAULT", Command.up(1.0)),
                    SingleDroneCommand("DEFAULT", Command.land())]
        self.assertEqual(expected, actual)

    def test_up_with_drone_name_should_return_correct_command(self):
        actual = generate_commands("""
            main() { DRONE1.takeoff(); DRONE1.up(1); DRONE1.land(); }
        """, drone_config_map={"DRONE1": DefaultDroneConfig()})
        expected = [SingleDroneCommand("DRONE1", Command.takeoff()),
                    SingleDroneCommand("DRONE1", Command.up(1)),
                    SingleDroneCommand("DRONE1", Command.land())]
        self.assertEqual(expected, actual)

    def test_up_with_incorrect_parameter_should_give_error(self):
        types = [Type.boolean(), Type.string(), Type.vector(), Type.drone(),
                 Type.list_of(Type.int()), Type.list_of(Type.list_of(Type.int()))]
        for type in types:
            with self.assertRaises(CompileError) as context:
                generate_commands("""
                    main() {{
                      {} a;
                      takeoff();
                      up(a);
                      land(); 
                    }}
                """.format(type.type_name))

            self.assertTrue("Expression {} should have type int or decimal, but is {}"
                            .format(Expression(type, type.default_value, ident="a"), type.type_name)
                            in str(context.exception))

    def test_down_with_int_parameter_should_return_correct_command(self):
        actual = generate_commands("""
            main() { takeoff(); down(1); land(); }
        """)
        expected = [SingleDroneCommand("DEFAULT", Command.takeoff()),
                    SingleDroneCommand("DEFAULT", Command.down(1)),
                    SingleDroneCommand("DEFAULT", Command.land())]
        self.assertEqual(expected, actual)

    def test_down_with_decimal_parameter_should_return_correct_command(self):
        actual = generate_commands("""
            main() { takeoff(); down(1.0); land(); }
        """)
        expected = [SingleDroneCommand("DEFAULT", Command.takeoff()),
                    SingleDroneCommand("DEFAULT", Command.down(1.0)),
                    SingleDroneCommand("DEFAULT", Command.land())]
        self.assertEqual(expected, actual)

    def test_down_with_drone_name_should_return_correct_command(self):
        actual = generate_commands("""
            main() { DRONE1.takeoff(); DRONE1.down(1); DRONE1.land(); }
        """, drone_config_map={"DRONE1": DefaultDroneConfig()})
        expected = [SingleDroneCommand("DRONE1", Command.takeoff()),
                    SingleDroneCommand("DRONE1", Command.down(1)),
                    SingleDroneCommand("DRONE1", Command.land())]
        self.assertEqual(expected, actual)

    def test_down_with_incorrect_parameter_should_give_error(self):
        types = [Type.boolean(), Type.string(), Type.vector(), Type.drone(),
                 Type.list_of(Type.int()), Type.list_of(Type.list_of(Type.int()))]
        for type in types:
            with self.assertRaises(CompileError) as context:
                generate_commands("""
                    main() {{
                      {} a;
                      takeoff();
                      down(a);
                      land(); 
                    }}
                """.format(type.type_name))

            self.assertTrue("Expression {} should have type int or decimal, but is {}"
                            .format(Expression(type, type.default_value, ident="a"), type.type_name)
                            in str(context.exception))

    def test_left_with_int_parameter_should_return_correct_command(self):
        actual = generate_commands("""
            main() { takeoff(); left(1); land(); }
        """)
        expected = [SingleDroneCommand("DEFAULT", Command.takeoff()),
                    SingleDroneCommand("DEFAULT", Command.left(1)),
                    SingleDroneCommand("DEFAULT", Command.land())]
        self.assertEqual(expected, actual)

    def test_left_with_decimal_parameter_should_return_correct_command(self):
        actual = generate_commands("""
            main() { takeoff(); left(1.0); land(); }
        """)
        expected = [SingleDroneCommand("DEFAULT", Command.takeoff()),
                    SingleDroneCommand("DEFAULT", Command.left(1.0)),
                    SingleDroneCommand("DEFAULT", Command.land())]
        self.assertEqual(expected, actual)

    def test_left_with_drone_name_should_return_correct_command(self):
        actual = generate_commands("""
            main() { DRONE1.takeoff(); DRONE1.left(1); DRONE1.land(); }
        """, drone_config_map={"DRONE1": DefaultDroneConfig()})
        expected = [SingleDroneCommand("DRONE1", Command.takeoff()),
                    SingleDroneCommand("DRONE1", Command.left(1)),
                    SingleDroneCommand("DRONE1", Command.land())]
        self.assertEqual(expected, actual)

    def test_left_with_incorrect_parameter_should_give_error(self):
        types = [Type.boolean(), Type.string(), Type.vector(), Type.drone(),
                 Type.list_of(Type.int()), Type.list_of(Type.list_of(Type.int()))]
        for type in types:
            with self.assertRaises(CompileError) as context:
                generate_commands("""
                    main() {{
                      {} a;
                      takeoff();
                      left(a);
                      land(); 
                    }}
                """.format(type.type_name))

            self.assertTrue("Expression {} should have type int or decimal, but is {}"
                            .format(Expression(type, type.default_value, ident="a"), type.type_name)
                            in str(context.exception))

    def test_right_with_int_parameter_should_return_correct_command(self):
        actual = generate_commands("""
            main() { takeoff(); right(1); land(); }
        """)
        expected = [SingleDroneCommand("DEFAULT", Command.takeoff()),
                    SingleDroneCommand("DEFAULT", Command.right(1)),
                    SingleDroneCommand("DEFAULT", Command.land())]
        self.assertEqual(expected, actual)

    def test_right_with_decimal_parameter_should_return_correct_command(self):
        actual = generate_commands("""
            main() { takeoff(); right(1.0); land(); }
        """)
        expected = [SingleDroneCommand("DEFAULT", Command.takeoff()),
                    SingleDroneCommand("DEFAULT", Command.right(1.0)),
                    SingleDroneCommand("DEFAULT", Command.land())]
        self.assertEqual(expected, actual)

    def test_right_with_drone_name_should_return_correct_command(self):
        actual = generate_commands("""
            main() { DRONE1.takeoff(); DRONE1.right(1); DRONE1.land(); }
        """, drone_config_map={"DRONE1": DefaultDroneConfig()})
        expected = [SingleDroneCommand("DRONE1", Command.takeoff()),
                    SingleDroneCommand("DRONE1", Command.right(1)),
                    SingleDroneCommand("DRONE1", Command.land())]
        self.assertEqual(expected, actual)

    def test_right_with_incorrect_parameter_should_give_error(self):
        types = [Type.boolean(), Type.string(), Type.vector(), Type.drone(),
                 Type.list_of(Type.int()), Type.list_of(Type.list_of(Type.int()))]
        for type in types:
            with self.assertRaises(CompileError) as context:
                generate_commands("""
                    main() {{
                      {} a;
                      takeoff();
                      right(a);
                      land(); 
                    }}
                """.format(type.type_name))

            self.assertTrue("Expression {} should have type int or decimal, but is {}"
                            .format(Expression(type, type.default_value, ident="a"), type.type_name)
                            in str(context.exception))

    def test_forward_with_int_parameter_should_return_correct_command(self):
        actual = generate_commands("""
            main() { takeoff(); forward(1); land(); }
        """)
        expected = [SingleDroneCommand("DEFAULT", Command.takeoff()),
                    SingleDroneCommand("DEFAULT", Command.forward(1)),
                    SingleDroneCommand("DEFAULT", Command.land())]
        self.assertEqual(expected, actual)

    def test_forward_with_decimal_parameter_should_return_correct_command(self):
        actual = generate_commands("""
            main() { takeoff(); forward(1.0); land(); }
        """)
        expected = [SingleDroneCommand("DEFAULT", Command.takeoff()),
                    SingleDroneCommand("DEFAULT", Command.forward(1.0)),
                    SingleDroneCommand("DEFAULT", Command.land())]
        self.assertEqual(expected, actual)

    def test_forward_with_drone_name_should_return_correct_command(self):
        actual = generate_commands("""
            main() { DRONE1.takeoff(); DRONE1.forward(1); DRONE1.land(); }
        """, drone_config_map={"DRONE1": DefaultDroneConfig()})
        expected = [SingleDroneCommand("DRONE1", Command.takeoff()),
                    SingleDroneCommand("DRONE1", Command.forward(1)),
                    SingleDroneCommand("DRONE1", Command.land())]
        self.assertEqual(expected, actual)

    def test_forward_with_incorrect_parameter_should_give_error(self):
        types = [Type.boolean(), Type.string(), Type.vector(), Type.drone(),
                 Type.list_of(Type.int()), Type.list_of(Type.list_of(Type.int()))]
        for type in types:
            with self.assertRaises(CompileError) as context:
                generate_commands("""
                    main() {{
                      {} a;
                      takeoff();
                      forward(a);
                      land(); 
                    }}
                """.format(type.type_name))

            self.assertTrue("Expression {} should have type int or decimal, but is {}"
                            .format(Expression(type, type.default_value, ident="a"), type.type_name)
                            in str(context.exception))

    def test_backward_with_int_parameter_should_return_correct_command(self):
        actual = generate_commands("""
            main() { takeoff(); backward(1); land(); }
        """)
        expected = [SingleDroneCommand("DEFAULT", Command.takeoff()),
                    SingleDroneCommand("DEFAULT", Command.backward(1)),
                    SingleDroneCommand("DEFAULT", Command.land())]
        self.assertEqual(expected, actual)

    def test_backward_with_decimal_parameter_should_return_correct_command(self):
        actual = generate_commands("""
            main() { takeoff(); backward(1.0); land(); }
        """)
        expected = [SingleDroneCommand("DEFAULT", Command.takeoff()),
                    SingleDroneCommand("DEFAULT", Command.backward(1.0)),
                    SingleDroneCommand("DEFAULT", Command.land())]
        self.assertEqual(expected, actual)

    def test_backward_with_drone_name_should_return_correct_command(self):
        actual = generate_commands("""
            main() { DRONE1.takeoff(); DRONE1.backward(1); DRONE1.land(); }
        """, drone_config_map={"DRONE1": DefaultDroneConfig()})
        expected = [SingleDroneCommand("DRONE1", Command.takeoff()),
                    SingleDroneCommand("DRONE1", Command.backward(1)),
                    SingleDroneCommand("DRONE1", Command.land())]
        self.assertEqual(expected, actual)

    def test_backward_with_incorrect_parameter_should_give_error(self):
        types = [Type.boolean(), Type.string(), Type.vector(), Type.drone(),
                 Type.list_of(Type.int()), Type.list_of(Type.list_of(Type.int()))]
        for type in types:
            with self.assertRaises(CompileError) as context:
                generate_commands("""
                    main() {{
                      {} a;
                      takeoff();
                      backward(a);
                      land(); 
                    }}
                """.format(type.type_name))

            self.assertTrue("Expression {} should have type int or decimal, but is {}"
                            .format(Expression(type, type.default_value, ident="a"), type.type_name)
                            in str(context.exception))

    def test_rotate_left_with_int_parameter_should_return_correct_command(self):
        actual = generate_commands("""
            main() { takeoff(); rotate_left(1); land(); }
        """)
        expected = [SingleDroneCommand("DEFAULT", Command.takeoff()),
                    SingleDroneCommand("DEFAULT", Command.rotate_left(1)),
                    SingleDroneCommand("DEFAULT", Command.land())]
        self.assertEqual(expected, actual)

    def test_rotate_left_with_decimal_parameter_should_return_correct_command(self):
        actual = generate_commands("""
            main() { takeoff(); rotate_left(1.0); land(); }
        """)
        expected = [SingleDroneCommand("DEFAULT", Command.takeoff()),
                    SingleDroneCommand("DEFAULT", Command.rotate_left(1.0)),
                    SingleDroneCommand("DEFAULT", Command.land())]
        self.assertEqual(expected, actual)

    def test_rotate_left_with_drone_name_should_return_correct_command(self):
        actual = generate_commands("""
            main() { DRONE1.takeoff(); DRONE1.rotate_left(1); DRONE1.land(); }
        """, drone_config_map={"DRONE1": DefaultDroneConfig()})
        expected = [SingleDroneCommand("DRONE1", Command.takeoff()),
                    SingleDroneCommand("DRONE1", Command.rotate_left(1)),
                    SingleDroneCommand("DRONE1", Command.land())]
        self.assertEqual(expected, actual)

    def test_rotate_left_with_incorrect_parameter_should_give_error(self):
        types = [Type.boolean(), Type.string(), Type.vector(), Type.drone(),
                 Type.list_of(Type.int()), Type.list_of(Type.list_of(Type.int()))]
        for type in types:
            with self.assertRaises(CompileError) as context:
                generate_commands("""
                    main() {{
                      {} a;
                      takeoff();
                      rotate_left(a);
                      land(); 
                    }}
                """.format(type.type_name))

            self.assertTrue("Expression {} should have type int or decimal, but is {}"
                            .format(Expression(type, type.default_value, ident="a"), type.type_name)
                            in str(context.exception))

    def test_rotate_right_with_int_parameter_should_return_correct_command(self):
        actual = generate_commands("""
            main() { takeoff(); rotate_right(1); land(); }
        """)
        expected = [SingleDroneCommand("DEFAULT", Command.takeoff()),
                    SingleDroneCommand("DEFAULT", Command.rotate_right(1)),
                    SingleDroneCommand("DEFAULT", Command.land())]
        self.assertEqual(expected, actual)

    def test_rotate_right_with_decimal_parameter_should_return_correct_command(self):
        actual = generate_commands("""
            main() { takeoff(); rotate_right(1.0); land(); }
        """)
        expected = [SingleDroneCommand("DEFAULT", Command.takeoff()),
                    SingleDroneCommand("DEFAULT", Command.rotate_right(1)),
                    SingleDroneCommand("DEFAULT", Command.land())]
        self.assertEqual(expected, actual)

    def test_rotate_right_with_drone_name_should_return_correct_command(self):
        actual = generate_commands("""
            main() { DRONE1.takeoff(); DRONE1.rotate_right(1); DRONE1.land(); }
        """, drone_config_map={"DRONE1": DefaultDroneConfig()})
        expected = [SingleDroneCommand("DRONE1", Command.takeoff()),
                    SingleDroneCommand("DRONE1", Command.rotate_right(1)),
                    SingleDroneCommand("DRONE1", Command.land())]
        self.assertEqual(expected, actual)

    def test_rotate_right_with_incorrect_parameter_should_give_error(self):
        types = [Type.boolean(), Type.string(), Type.vector(), Type.drone(),
                 Type.list_of(Type.int()), Type.list_of(Type.list_of(Type.int()))]
        for type in types:
            with self.assertRaises(CompileError) as context:
                generate_commands("""
                    main() {{
                      {} a;
                      takeoff();
                      rotate_right(a);
                      land(); 
                    }}
                """.format(type.type_name))

            self.assertTrue("Expression {} should have type int or decimal, but is {}"
                            .format(Expression(type, type.default_value, ident="a"), type.type_name)
                            in str(context.exception))

    def test_wait_with_int_parameter_should_return_correct_command(self):
        actual = generate_commands("""
            main() { wait(1); }
        """)
        expected = [SingleDroneCommand("DEFAULT", Command.wait(1))]
        self.assertEqual(expected, actual)

    def test_wait_with_decimal_parameter_should_return_correct_command(self):
        actual = generate_commands("""
            main() { wait(1.0); }
        """)
        expected = [SingleDroneCommand("DEFAULT", Command.wait(1.0))]
        self.assertEqual(expected, actual)

    def test_wait_with_drone_name_should_return_correct_command(self):
        actual = generate_commands("""
            main() { DRONE1.wait(1); }
        """, drone_config_map={"DRONE1": DefaultDroneConfig()})
        expected = [SingleDroneCommand("DRONE1", Command.wait(1))]
        self.assertEqual(expected, actual)

    def test_wait_with_incorrect_parameter_should_give_error(self):
        types = [Type.boolean(), Type.string(), Type.vector(), Type.drone(),
                 Type.list_of(Type.int()), Type.list_of(Type.list_of(Type.int()))]
        for type in types:
            with self.assertRaises(CompileError) as context:
                generate_commands("""
                    main() {{
                      {} a;
                      wait(a); 
                    }}
                """.format(type.type_name))

            self.assertTrue("Expression {} should have type int or decimal, but is {}"
                            .format(Expression(type, type.default_value, ident="a"), type.type_name)
                            in str(context.exception))

    def test_multiple_commands_should_return_correct_command(self):
        actual = generate_commands("""
            main() {
             takeoff();
             up(1);
             down(2);
             left(3);
             right(4);
             forward(5);
             backward(6);
             rotate_left(7);
             rotate_right(8);
             wait(9);
             land(); 
            }
        """)
        expected = [SingleDroneCommand("DEFAULT", Command.takeoff()),
                    SingleDroneCommand("DEFAULT", Command.up(1)),
                    SingleDroneCommand("DEFAULT", Command.down(2)),
                    SingleDroneCommand("DEFAULT", Command.left(3)),
                    SingleDroneCommand("DEFAULT", Command.right(4)),
                    SingleDroneCommand("DEFAULT", Command.forward(5)),
                    SingleDroneCommand("DEFAULT", Command.backward(6)),
                    SingleDroneCommand("DEFAULT", Command.rotate_left(7)),
                    SingleDroneCommand("DEFAULT", Command.rotate_right(8)),
                    SingleDroneCommand("DEFAULT", Command.wait(9)),
                    SingleDroneCommand("DEFAULT", Command.land())]
        self.assertEqual(expected, actual)


class MovementGetDroneNameTest(unittest.TestCase):
    def test_movement_with_no_name_specified_if_only_one_drone_should_use_that_drone(self):
        drone_config_map = {"THE_ONE": DefaultDroneConfig()}
        actual = generate_commands("""
            main() { takeoff(); land(); }
        """, drone_config_map=drone_config_map)
        expected = [SingleDroneCommand("THE_ONE", Command.takeoff()),
                    SingleDroneCommand("THE_ONE", Command.land())]
        self.assertEqual(expected, actual)

    def test_movement_with_no_name_specified_if_multiple_drones_should_give_error(self):
        drone_config_map = {"DRONE1": DefaultDroneConfig(), "DRONE2": DefaultDroneConfig()}
        with self.assertRaises(CompileError) as context:
            generate_commands("""
                main() { takeoff(); land(); }
            """, drone_config_map=drone_config_map)
        self.assertTrue("Drone should be specified if there are multiple drones in config"
                        in str(context.exception))

    def test_movement_with_undefined_name_specified_should_give_error(self):
        drone_config_map = {"DRONE1": DefaultDroneConfig()}
        with self.assertRaises(CompileError) as context:
            generate_commands("""
                main() { DRONE2.takeoff(); DRONE2.land(); }
            """, drone_config_map=drone_config_map)
        self.assertTrue("Identifier DRONE2 has not been declared"
                        in str(context.exception))

    def test_movement_with_null_drone_should_give_error(self):
        with self.assertRaises(CompileError) as context:
            generate_commands("""
                main() { drone d; d.takeoff(); d.land(); }
            """)
        self.assertTrue("Drone has not been assigned"
                        in str(context.exception))

    def test_movement_with_other_type_should_give_error(self):
        types = [Type.int(), Type.decimal(), Type.string(), Type.boolean(), Type.vector(),
                 Type.list_of(Type.int()), Type.list_of(Type.decimal()), Type.list_of(Type.list_of(Type.int()))]
        for type in types:
            with self.assertRaises(CompileError) as context:
                generate_commands("""
                    main() {{ {} a; a.takeoff(); a.land(); }}
                """.format(type))
            self.assertTrue("Expression {} should have type drone, but is {}"
                            .format(Expression(type, type.default_value, ident="a"), type.type_name)
                            in str(context.exception))



import unittest
from unittest.mock import patch, Mock

from xdrone.command_generators import generate_commands, generate_commands_with_config
from xdrone.safety_checker.boundary_checker import BoundaryChecker
from xdrone.shared.boundary_config import BoundaryConfig
from xdrone.shared.command import Command, SingleDroneCommand
from xdrone.shared.compile_error import CompileError, XDroneSyntaxError
from xdrone.shared.drone_config import DroneConfig
from xdrone.shared.safety_check_error import SafetyCheckError


class GenerateCommandsTest(unittest.TestCase):
    def test_basic_program(self):
        commands = "main() {takeoff(); land();}"
        expected = [SingleDroneCommand("DEFAULT", Command.takeoff()), SingleDroneCommand("DEFAULT", Command.land())]
        self.assertEqual(expected, generate_commands(commands))

    def test_syntax_error_should_be_reported(self):
        commands = "main() {takeoff() land();}"
        with self.assertRaises(XDroneSyntaxError) as context:
            generate_commands(commands)
        self.assertTrue("missing ';' at 'land'" in str(context.exception))

    def test_all_syntax_errors_should_be_reported(self):
        commands = "main() {takeoff() land()}"
        with self.assertRaises(XDroneSyntaxError) as context:
            generate_commands(commands)
        self.assertTrue("missing ';' at 'land'" in str(context.exception))
        self.assertTrue("missing ';' at '}'" in str(context.exception))

    def test_compile_error_should_be_reported(self):
        commands = "main() {takeoff(); up(\"abc\"); land();}"
        with self.assertRaises(CompileError) as context:
            generate_commands(commands)
        self.assertTrue("should have type int or decimal, but is string" in str(context.exception))

    def test_if_not_given_state_updater_should_use_default_to_update_state(self):
        commands = "main() {takeoff(); land();}"
        generate_commands(commands, boundary_checker=BoundaryChecker(BoundaryConfig(max_seconds=10, max_z_meters=1)))

    def test_if_given_state_updater_should_use_it_to_update_state(self):
        commands = "main() {takeoff(); land();}"
        with self.assertRaises(SafetyCheckError) as context:
            generate_commands(commands, drone_config_map={"DEFAULT": DroneConfig(init_position=(0, 0, 0),
                                                                                 speed_mps=1, rotate_speed_dps=90,
                                                                                 takeoff_height_meters=10)},
                              boundary_checker=BoundaryChecker(BoundaryConfig(max_seconds=10, max_z_meters=1)))
        self.assertTrue("Drone 'DEFAULT': the z coordinate 10 will go beyond its upper limit 1"
                        in str(context.exception))

    def test_if_not_given_boundary_checker_should_use_default_to_check_safety(self):
        commands = "main() {takeoff(); wait(1000); up(1000); land();}"
        generate_commands(commands)

    def test_if_given_boundary_checker_should_use_it_to_check_safety(self):
        commands = "main() {takeoff(); up(1000); land();}"
        with self.assertRaises(SafetyCheckError) as context:
            generate_commands(commands,
                              boundary_checker=BoundaryChecker(BoundaryConfig(max_seconds=10000, max_z_meters=1)))
        self.assertTrue("Drone 'DEFAULT': the z coordinate 1001 will go beyond its upper limit 1"
                        in str(context.exception))


class GenerateCommandsWithConfig(unittest.TestCase):

    @patch('xdrone.command_generators.ConfigParser.parse')
    @patch('xdrone.command_generators.generate_commands')
    @patch('xdrone.command_generators.CollisionChecker')
    @patch('xdrone.command_generators.BoundaryChecker')
    def test_generate_commands_with_config(self, mock_boundary_checker, mock_collision_checker,
                                           mock_generate_commands, mock_parse):
        program, config_json, has_checks, save_report = Mock(), Mock(), Mock(), Mock()
        drone_config_map, boundary_config, collision_config = Mock(), Mock(), Mock()
        boundary_checker = Mock()
        collision_checker = Mock()
        generated_commands = Mock()
        mock_parse.return_value = (drone_config_map, boundary_config, collision_config)
        mock_generate_commands.return_value = generated_commands
        mock_collision_checker.return_value = collision_checker
        mock_boundary_checker.return_value = boundary_checker

        self.assertEqual((generated_commands, drone_config_map, boundary_config),
                         generate_commands_with_config(program, config_json, has_checks, save_report))

        mock_generate_commands.assert_called_with(program, drone_config_map, boundary_checker,
                                                  collision_checker, has_checks, save_report)



import unittest

from xdrone.compiler.compiler_utils.drones import Drone
from xdrone.compiler.compiler_utils.expressions import Expression, AbstractExpression, Identifier, ListElem, VectorElem
from xdrone.compiler.compiler_utils.type import Type
from xdrone.shared.drone_config import DefaultDroneConfig


class TestAbstractExpression(unittest.TestCase):
    def test_to_expression(self):
        with self.assertRaises(NotImplementedError) as context:
            AbstractExpression().to_expression()


class TestExpression(unittest.TestCase):
    def test_to_expression_should_equal_self(self):
        expressions = [Expression(Type.int(), 1), Expression(Type.decimal(), 1.1), Expression(Type.boolean(), False),
                       Expression(Type.vector(), [1.1, 2.2, -1.1]), Expression(Type.list_of(Type.int()), [1, 2, 3, 4]),
                       Expression(Type.list_of(Type.int()), []),
                       Expression(Type.list_of(Type.decimal()), [1.0, 2.0, 3.0, 4.0]),
                       Expression(Type.list_of(Type.list_of(Type.vector())), [[[1.1, 2.2, -1.1], [1, 2, -1]]]),
                       Expression(Type.int(), 1, "a"), Expression(Type.decimal(), 1.1, "a"),
                       Expression(Type.boolean(), False, "a"),
                       Expression(Type.vector(), [1.1, 2.2, -1.1], "a"),
                       Expression(Type.drone(), Drone("DRONE", DefaultDroneConfig()), "a"),
                       Expression(Type.list_of(Type.int()), [1, 2, 3, 4], "a"),
                       Expression(Type.list_of(Type.int()), [], "a"),
                       Expression(Type.list_of(Type.decimal()), [1.0, 2.0, 3.0, 4.0], "a"),
                       Expression(Type.list_of(Type.list_of(Type.vector())), [[[1.1, 2.2, -1.1], [1, 2, -1]]], "a")]
        for expression in expressions:
            self.assertEqual(expression, expression.to_expression())

    def test_eq(self):
        expressions1 = [None,
                        Expression(Type.int(), 1), Expression(Type.decimal(), 1.1), Expression(Type.boolean(), False),
                        Expression(Type.vector(), [1.1, 2.2, -1.1]), Expression(Type.list_of(Type.int()), [1, 2, 3, 4]),
                        Expression(Type.list_of(Type.int()), []),
                        Expression(Type.list_of(Type.decimal()), [1.0, 2.0, 3.0, 4.0]),
                        Expression(Type.list_of(Type.list_of(Type.vector())), [[[1.1, 2.2, -1.1], [1, 2, -1]]]),
                        Expression(Type.int(), 1, "a"), Expression(Type.decimal(), 1.1, "a"),
                        Expression(Type.boolean(), False, "a"),
                        Expression(Type.vector(), [1.1, 2.2, -1.1], "a"),
                        Expression(Type.drone(), Drone("DRONE", DefaultDroneConfig()), "a"),
                        Expression(Type.list_of(Type.int()), [1, 2, 3, 4], "a"),
                        Expression(Type.list_of(Type.int()), [], "a"),
                        Expression(Type.list_of(Type.decimal()), [1.0, 2.0, 3.0, 4.0], "a"),
                        Expression(Type.list_of(Type.list_of(Type.vector())), [[[1.1, 2.2, -1.1], [1, 2, -1]]], "a")]
        expressions2 = [None,
                        Expression(Type.int(), 1), Expression(Type.decimal(), 1.1), Expression(Type.boolean(), False),
                        Expression(Type.vector(), [1.1, 2.2, -1.1]), Expression(Type.list_of(Type.int()), [1, 2, 3, 4]),
                        Expression(Type.list_of(Type.int()), []),
                        Expression(Type.list_of(Type.decimal()), [1.0, 2.0, 3.0, 4.0]),
                        Expression(Type.list_of(Type.list_of(Type.vector())), [[[1.1, 2.2, -1.1], [1, 2, -1]]]),
                        Expression(Type.int(), 1, "a"), Expression(Type.decimal(), 1.1, "a"),
                        Expression(Type.boolean(), False, "a"),
                        Expression(Type.vector(), [1.1, 2.2, -1.1], "a"),
                        Expression(Type.drone(), Drone("DRONE", DefaultDroneConfig()), "a"),
                        Expression(Type.list_of(Type.int()), [1, 2, 3, 4], "a"),
                        Expression(Type.list_of(Type.int()), [], "a"),
                        Expression(Type.list_of(Type.decimal()), [1.0, 2.0, 3.0, 4.0], "a"),
                        Expression(Type.list_of(Type.list_of(Type.vector())), [[[1.1, 2.2, -1.1], [1, 2, -1]]], "a")]
        for i in range(len(expressions1)):
            for j in range(len(expressions2)):
                if i == j:
                    self.assertEqual(expressions1[i], expressions2[j])
                else:
                    self.assertNotEqual(expressions1[i], expressions2[j])

    def test_str(self):
        self.assertEqual("Expression: { type: int, value: 1, ident: None }", str(Expression(Type.int(), 1)))
        self.assertEqual("Expression: { type: decimal, value: 1.1, ident: None }", str(Expression(Type.decimal(), 1.1)))
        self.assertEqual("Expression: { type: boolean, value: False, ident: None }",
                         str(Expression(Type.boolean(), False)))
        self.assertEqual("Expression: { type: vector, value: [1.1, 2.2, -1.1], ident: None }",
                         str(Expression(Type.vector(), [1.1, 2.2, -1.1])))
        self.assertEqual("Expression: { type: list[int], value: [1, 2, 3, 4], ident: None }",
                         str(Expression(Type.list_of(Type.int()), [1, 2, 3, 4])))
        self.assertEqual("Expression: { type: list[int], value: [], ident: None }",
                         str(Expression(Type.list_of(Type.int()), [])))
        self.assertEqual(
            "Expression: { type: list[list[vector]], value: [[[1.1, 2.2, -1.1], [1, 2, -1]]], ident: None }",
            str(Expression(Type.list_of(Type.list_of(Type.vector())), [[[1.1, 2.2, -1.1], [1, 2, -1]]])))

        self.assertEqual("Expression: { type: int, value: 1, ident: a }", str(Expression(Type.int(), 1, "a")))
        self.assertEqual("Expression: { type: decimal, value: 1.1, ident: a }",
                         str(Expression(Type.decimal(), 1.1, "a")))
        self.assertEqual("Expression: { type: boolean, value: False, ident: a }",
                         str(Expression(Type.boolean(), False, "a")))
        self.assertEqual("Expression: { type: vector, value: [1.1, 2.2, -1.1], ident: a }",
                         str(Expression(Type.vector(), [1.1, 2.2, -1.1], "a")))
        self.assertEqual("Expression: {{ type: drone, value: {}, ident: a }}"
                         .format(Drone("DRONE", DefaultDroneConfig())),
                         str(Expression(Type.drone(), Drone("DRONE", DefaultDroneConfig()), "a")))
        self.assertEqual("Expression: { type: list[int], value: [1, 2, 3, 4], ident: a }",
                         str(Expression(Type.list_of(Type.int()), [1, 2, 3, 4], "a")))
        self.assertEqual("Expression: { type: list[int], value: [], ident: a }",
                         str(Expression(Type.list_of(Type.int()), [], "a")))
        self.assertEqual("Expression: { type: list[list[vector]], value: [[[1.1, 2.2, -1.1], [1, 2, -1]]], ident: a }",
                         str(Expression(Type.list_of(Type.list_of(Type.vector())), [[[1.1, 2.2, -1.1], [1, 2, -1]]],
                                        "a")))


class TestIdentifier(unittest.TestCase):
    def test_to_expression_should_return_correct_value(self):
        expressions = [None, Expression(Type.int(), 1, "a"), Expression(Type.decimal(), 1.1, "a"),
                       Expression(Type.boolean(), False, "a"),
                       Expression(Type.vector(), [1.1, 2.2, -1.1], "a"),
                       Expression(Type.drone(), Drone("DRONE", DefaultDroneConfig()), "a"),
                       Expression(Type.list_of(Type.int()), [1, 2, 3, 4], "a"),
                       Expression(Type.list_of(Type.int()), [], "a"),
                       Expression(Type.list_of(Type.decimal()), [1.0, 2.0, 3.0, 4.0], "a"),
                       Expression(Type.list_of(Type.list_of(Type.vector())), [[[1.1, 2.2, -1.1], [1, 2, -1]]], "a")]
        for expression in expressions:
            self.assertEqual(expression, Identifier("a", expression).to_expression())

    def test_eq(self):
        identifiers1 = [None, Identifier("a", None), Identifier("a", Expression(Type.int(), 1, "a")),
                        Identifier("a", Expression(Type.decimal(), 1.1, "a")),
                        Identifier("b", Expression(Type.int(), 1, "b")),
                        Identifier("b", Expression(Type.list_of(Type.int()), [], "b"))]
        identifiers2 = [None, Identifier("a", None), Identifier("a", Expression(Type.int(), 1, "a")),
                        Identifier("a", Expression(Type.decimal(), 1.1, "a")),
                        Identifier("b", Expression(Type.int(), 1, "b")),
                        Identifier("b", Expression(Type.list_of(Type.int()), [], "b"))]
        for i in range(len(identifiers1)):
            for j in range(len(identifiers2)):
                if i == j:
                    self.assertEqual(identifiers1[i], identifiers2[j])
                else:
                    self.assertNotEqual(identifiers1[i], identifiers2[j])

    def test_str(self):
        self.assertEqual("Identifier: { ident: a, expression: None }", str(Identifier("a", None)))
        self.assertEqual("Identifier: { ident: a1, expression: Expression: { type: int, value: 1, ident: a1 } }",
                         str(Identifier("a1", Expression(Type.int(), 1, ident="a1"))))


class TestListElem(unittest.TestCase):
    def test_container_not_list_should_give_error(self):
        none_list_exprs = [Expression(Type.int(), 1, "a"), Expression(Type.decimal(), 1.0, "a"),
                           Expression(Type.boolean(), True, "a"), Expression(Type.string(), "abc", "a"),
                           Expression(Type.vector(), [1.0, 1.0, 1.0], "a"),
                           Expression(Type.drone(), Drone("DRONE", DefaultDroneConfig()), "a")]
        for none_list_expr in none_list_exprs:
            with self.assertRaises(AssertionError) as context:
                ListElem("a", none_list_expr, 0)

    def test_to_expression_should_return_correct_value(self):
        expression = Expression(Type.list_of(Type.decimal()), [1.0, 2.0, 3.0, 4.0], "a")
        expected = Expression(Type.decimal(), 1.0, "a[0]")
        self.assertEqual(expected, ListElem("a", expression, 0).to_expression())

        expression = Expression(Type.list_of(Type.list_of(Type.vector())), [[[1.1, 2.2, -1.1], [1, 2, -1]]], "a")
        expected = Expression(Type.list_of(Type.vector()), [[1.1, 2.2, -1.1], [1, 2, -1]], "a[0]")
        self.assertEqual(expected, ListElem("a", expression, 0).to_expression())

        expression = Expression(Type.list_of(Type.boolean()), [True, False], "b[1]")
        expected = Expression(Type.boolean(), False, "b[1][1]")
        self.assertEqual(expected, ListElem("b[1]", expression, 1).to_expression())

    def test_eq(self):
        list_elems1 = [None,
                       ListElem("a", Expression(Type.list_of(Type.decimal()), [1.0, 2.0, 3.0, 4.0], "a"), 0),
                       ListElem("b", Expression(Type.list_of(Type.decimal()), [1.0, 2.0, 3.0, 4.0], "b"), 0),
                       ListElem("a", Expression(Type.list_of(Type.decimal()), [1.0, 2.0, 3.0], "a"), 0),
                       ListElem("a", Expression(Type.list_of(Type.decimal()), [1.0, 2.0, 3.0, 4.0], "a"), 1),
                       ListElem("a", Expression(Type.list_of(Type.list_of(Type.int())), [[[1], [1]]], "a"), 0),
                       ListElem("b[1]", Expression(Type.list_of(Type.boolean()), [True, False], "b[1]"), 0)]
        list_elems2 = [None,
                       ListElem("a", Expression(Type.list_of(Type.decimal()), [1.0, 2.0, 3.0, 4.0], "a"), 0),
                       ListElem("b", Expression(Type.list_of(Type.decimal()), [1.0, 2.0, 3.0, 4.0], "b"), 0),
                       ListElem("a", Expression(Type.list_of(Type.decimal()), [1.0, 2.0, 3.0], "a"), 0),
                       ListElem("a", Expression(Type.list_of(Type.decimal()), [1.0, 2.0, 3.0, 4.0], "a"), 1),
                       ListElem("a", Expression(Type.list_of(Type.list_of(Type.int())), [[[1], [1]]], "a"), 0),
                       ListElem("b[1]", Expression(Type.list_of(Type.boolean()), [True, False], "b[1]"), 0)]
        for i in range(len(list_elems1)):
            for j in range(len(list_elems2)):
                if i == j:
                    self.assertEqual(list_elems1[i], list_elems2[j])
                else:
                    self.assertNotEqual(list_elems1[i], list_elems2[j])

    def test_str(self):
        self.assertEqual("ListElem: { ident: a, " +
                         "container: Expression: { type: list[int], value: [1, 2, 3, 4], ident: a }, " +
                         "index: 0, expression: Expression: { type: int, value: 1, ident: a[0] } }",
                         str(ListElem("a", Expression(Type.list_of(Type.int()), [1, 2, 3, 4], "a"), 0)))


class TestVectorElem(unittest.TestCase):
    def test_container_not_vector_should_give_error(self):
        none_vector_exprs = [Expression(Type.int(), 1, "a"), Expression(Type.decimal(), 1.0, "a"),
                             Expression(Type.boolean(), True, "a"), Expression(Type.string(), "abc", "a"),
                             Expression(Type.drone(), Drone("DRONE", DefaultDroneConfig()), "a"),
                             Expression(Type.list_of(Type.decimal()), [1.0, 1.0, 1.0], "a"),
                             Expression(Type.list_of(Type.list_of(Type.decimal())), [[1.0, 1.0, 1.0]], "a")]
        for none_vector_expr in none_vector_exprs:
            with self.assertRaises(AssertionError) as context:
                VectorElem("a", none_vector_expr, 0)

    def test_to_expression_should_return_correct_value(self):
        expression = Expression(Type.vector(), [1.0, 2.0, 3.0], "a")
        expected = Expression(Type.decimal(), 1.0)
        self.assertEqual(expected, VectorElem("a", expression, 0).to_expression())

        expression = Expression(Type.vector(), [1.0, 2.0, 3.0], "b[1]")
        expected = Expression(Type.decimal(), 3.0)
        self.assertEqual(expected, VectorElem("b[1]", expression, 2).to_expression())

    def test_eq(self):
        vector_elems1 = [None,
                         VectorElem("a", Expression(Type.vector(), [1.0, 2.0, 3.0], "a"), 0),
                         VectorElem("b", Expression(Type.vector(), [1.0, 2.0, 3.0], "b"), 0),
                         VectorElem("a", Expression(Type.vector(), [2.0, 2.0, 3.0], "a"), 0),
                         VectorElem("a", Expression(Type.vector(), [1.0, 2.0, 3.0], "a"), 1),
                         VectorElem("b[1]", Expression(Type.vector(), [1.0, 2.0, 3.0], "a"), 0)]
        vector_elems2 = [None,
                         VectorElem("a", Expression(Type.vector(), [1.0, 2.0, 3.0], "a"), 0),
                         VectorElem("b", Expression(Type.vector(), [1.0, 2.0, 3.0], "b"), 0),
                         VectorElem("a", Expression(Type.vector(), [2.0, 2.0, 3.0], "a"), 0),
                         VectorElem("a", Expression(Type.vector(), [1.0, 2.0, 3.0], "a"), 1),
                         VectorElem("b[1]", Expression(Type.vector(), [1.0, 2.0, 3.0], "a"), 0)]
        for i in range(len(vector_elems1)):
            for j in range(len(vector_elems2)):
                if i == j:
                    self.assertEqual(vector_elems1[i], vector_elems2[j])
                else:
                    self.assertNotEqual(vector_elems1[i], vector_elems2[j])

    def test_str(self):
        self.assertEqual("ListElem: { ident: a, " +
                         "container: Expression: { type: vector, value: [1.0, 2.0, 3.0], ident: a }, " +
                         "index: 0, expression: Expression: { type: decimal, value: 1.0, ident: None } }",
                         str(VectorElem("a", Expression(Type.vector(), [1.0, 2.0, 3.0], "a"), 0)))



import unittest

from xdrone.compiler.compiler_utils.drones import Drone, NullDrone
from xdrone.shared.drone_config import DefaultDroneConfig, DroneConfig


class TestDrone(unittest.TestCase):
    def test_property(self):
        self.assertEqual("a", Drone("a", DefaultDroneConfig()).name)
        self.assertEqual(DefaultDroneConfig(), Drone("a", DefaultDroneConfig()).config)

    def test_str(self):
        self.assertEqual("Drone: {{ name: a, config: {} }}".format(DefaultDroneConfig())
                         , str(Drone("a", DefaultDroneConfig())))

    def test_eq(self):
        self.assertEqual(Drone("a", DefaultDroneConfig()), Drone("a", DefaultDroneConfig()))
        self.assertNotEqual(Drone("b", DefaultDroneConfig()), Drone("a", DefaultDroneConfig()))
        self.assertNotEqual(Drone("a", DroneConfig((1, 2, 3), 1, 90, 2)), Drone("a", DefaultDroneConfig()))
        self.assertNotEqual(None, Drone("a", DefaultDroneConfig()))


class TestNullDrone(unittest.TestCase):
    def test_property(self):
        self.assertEqual("null", NullDrone().name)
        self.assertEqual(DefaultDroneConfig(), NullDrone().config)

    def test_str(self):
        self.assertEqual("NullDrone: { }", str(NullDrone()))

    def test_eq(self):
        self.assertEqual(NullDrone(), NullDrone())
        self.assertNotEqual(Drone("a", DefaultDroneConfig()), NullDrone())
        self.assertNotEqual(Drone("null", DefaultDroneConfig()), NullDrone())
        self.assertNotEqual(None, NullDrone())



import unittest

from xdrone.command_generators import generate_commands
from xdrone.compiler.compiler_utils.expressions import Expression
from xdrone.compiler.compiler_utils.functions import FunctionTable, Function, Parameter
from xdrone.compiler.compiler_utils.symbol_table import SymbolTable
from xdrone.compiler.compiler_utils.type import Type
from xdrone.shared.command import Command, SingleDroneCommand
from xdrone.shared.compile_error import CompileError
from xdrone.shared.drone_config import DefaultDroneConfig


class FunctionDefinitionTest(unittest.TestCase):
    def test_define_functions_should_change_function_table(self):
        types = [Type.int(), Type.decimal(), Type.string(), Type.boolean(), Type.vector(), Type.drone(),
                 Type.list_of(Type.int()), Type.list_of(Type.decimal()), Type.list_of(Type.list_of(Type.int()))]
        for type in types:
            actual = FunctionTable()
            generate_commands("""
                function func() return {} {{}}
                procedure proc() {{}}
                main () {{}}
                """.format(type), function_table=actual)
            expected = FunctionTable()
            expected.store("func", Function("func", [], type, []))
            expected.store("proc", Function("proc", [], None, []))
            self.assertEqual(expected, actual)

    def test_define_functions_with_parameter_should_change_function_table(self):
        types = [Type.int(), Type.decimal(), Type.string(), Type.boolean(), Type.vector(), Type.drone(),
                 Type.list_of(Type.int()), Type.list_of(Type.decimal()), Type.list_of(Type.list_of(Type.int()))]
        for type1 in types:
            for type2 in types:
                actual = FunctionTable()
                generate_commands("""
                    function func({} a, {} b) return int {{}}
                    procedure proc({} a, {} b) {{}}
                    main () {{}}
                    """.format(type1, type2, type1, type2), function_table=actual)
                expected = FunctionTable()
                expected.store("func", Function("func", [Parameter("a", type1), Parameter("b", type2)], Type.int(), []))
                expected.store("proc", Function("proc", [Parameter("a", type1), Parameter("b", type2)], None, []))
                self.assertEqual(expected, actual)

    def test_define_functions_with_duplicated_parameter_should_give_error(self):
        types = [Type.int(), Type.decimal(), Type.string(), Type.boolean(), Type.vector(), Type.drone(),
                 Type.list_of(Type.int()), Type.list_of(Type.decimal()), Type.list_of(Type.list_of(Type.int()))]
        for type1 in types:
            for type2 in types:
                with self.assertRaises(CompileError) as context:
                    generate_commands("""
                        function func({} a, {} a) return int {{}}
                        main () {{}}
                        """.format(type1, type2, type1, type2))
                self.assertTrue("Parameter names are duplicated in ['a', 'a']" in str(context.exception))

                with self.assertRaises(CompileError) as context:
                    generate_commands("""
                        procedure proc({} a, {} a) {{}}
                        main () {{}}
                        """.format(type1, type2, type1, type2))
                self.assertTrue("Parameter names are duplicated in ['a', 'a']" in str(context.exception))

    def test_define_existed_function_should_give_error(self):
        lines = ["function func() return int {}", "procedure func() {}"]
        for line1 in lines:
            for line2 in lines:
                with self.assertRaises(CompileError) as context:
                    generate_commands("""
                        {}
                        {}
                        main () {{}}
                        """.format(line1, line2))

                self.assertTrue("Function or procedure func already defined" in str(context.exception))


class ReturnTest(unittest.TestCase):
    def test_return_correct_type_in_function_should_return_correct_value(self):
        types = [Type.int(), Type.decimal(), Type.string(), Type.boolean(), Type.vector(), Type.drone(),
                 Type.list_of(Type.int()), Type.list_of(Type.decimal()), Type.list_of(Type.list_of(Type.int()))]
        for type in types:
            actual = SymbolTable()
            generate_commands("""
                function func() return {} {{
                  {} a;
                  return a;
                }}
                main () {{
                  {} a <- func();
                }}
                """.format(type, type, type), symbol_table=actual)
            expected = SymbolTable()
            expected.store("a", Expression(type, type.default_value, ident="a"))
            self.assertEqual(expected, actual)

    def test_return_in_function_should_exit_early(self):
        commands = generate_commands("""
            function func() return int {
              return 1;
              up(100);
            }
            main () {
              int a <- func();
            }
            """)
        self.assertEqual([], commands)

    def test_return_wrong_type_in_function_should_give_error(self):
        types = [Type.int(), Type.decimal(), Type.string(), Type.boolean(), Type.vector(), Type.drone(),
                 Type.list_of(Type.int()), Type.list_of(Type.decimal()), Type.list_of(Type.list_of(Type.int()))]
        for t1 in types:
            for t2 in types:
                if t1 == t2:
                    continue
                with self.assertRaises(CompileError) as context:
                    generate_commands("""
                        function func() return {} {{
                          {} a;
                          return a;
                        }}
                        main () {{
                          {} a <- func();
                        }}
                        """.format(t1, t2, t1))
                self.assertTrue("Function func has returned type {}, but {} is returned"
                                .format(t1, t2)
                                in str(context.exception))

    def test_return_empty_in_function_should_give_error(self):
        types = [Type.int(), Type.decimal(), Type.string(), Type.boolean(), Type.vector(), Type.drone(),
                 Type.list_of(Type.int()), Type.list_of(Type.decimal()), Type.list_of(Type.list_of(Type.int()))]
        for type in types:
            with self.assertRaises(CompileError) as context:
                generate_commands("""
                    function func() return {} {{
                      return;
                    }}
                    main () {{
                      {} a <- func();
                    }}
                    """.format(type, type))
            self.assertTrue("Function func has returned type {}, but nothing is returned"
                            .format(type)
                            in str(context.exception))

    def test_return_not_exist_in_function_should_give_error(self):
        types = [Type.int(), Type.decimal(), Type.string(), Type.boolean(), Type.vector(), Type.drone(),
                 Type.list_of(Type.int()), Type.list_of(Type.decimal()), Type.list_of(Type.list_of(Type.int()))]
        for type in types:
            with self.assertRaises(CompileError) as context:
                generate_commands("""
                    function func() return {} {{
                    }}
                    main () {{
                      {} a <- func();
                    }}
                    """.format(type, type))
            self.assertTrue("Function func has returned type {}, but nothing is returned"
                            .format(type)
                            in str(context.exception))

    def test_return_empty_in_procedure_should_exit_early(self):
        commands = generate_commands("""
            procedure proc() {
              return;
              up(100);
            }
            main () {
              proc();
            }
            """)
        self.assertEqual([], commands)

    def test_do_not_return_in_procedure_should_success(self):
        commands = generate_commands("""
            procedure proc() {
            }
            main () {
              proc();
            }
            """)
        self.assertEqual([], commands)

    def test_return_value_in_procedure_should_give_error(self):
        types = [Type.int(), Type.decimal(), Type.string(), Type.boolean(), Type.vector(), Type.drone(),
                 Type.list_of(Type.int()), Type.list_of(Type.decimal()), Type.list_of(Type.list_of(Type.int()))]
        for type in types:
            with self.assertRaises(CompileError) as context:
                generate_commands("""
                    procedure proc() {{
                      {} a;
                      return a;
                    }}
                    main () {{
                      proc();
                    }}
                    """.format(type))
            self.assertTrue("Procedure proc should not return anything, but {} is returned"
                            .format(Expression(type, type.default_value, ident=None))
                            in str(context.exception))

    def test_return_in_main_should_give_error(self):
        types = [Type.int(), Type.decimal(), Type.string(), Type.boolean(), Type.vector(), Type.drone(),
                 Type.list_of(Type.int()), Type.list_of(Type.decimal()), Type.list_of(Type.list_of(Type.int()))]
        for type in types:
            with self.assertRaises(CompileError) as context:
                generate_commands("""
                    main () {{
                      {} a;
                      return a;
                    }}
                    """.format(type))
            self.assertTrue("Cannot return in the Main function" in str(context.exception))

    def test_return_empty_in_main_should_give_error(self):
        with self.assertRaises(CompileError) as context:
            generate_commands("""
                main () {
                  return;
                }
                """)
        self.assertTrue("Cannot return in the Main function" in str(context.exception))


class ProcedureCallTest(unittest.TestCase):
    def test_call_procedure_should_run_commands(self):
        commands = generate_commands("""
            procedure proc(int i, int j) {
              up(i + j);
              down(i);
            }
            main () {
              takeoff();
              proc(100, 200);
              land();
            }
            """)
        expected_commands = [SingleDroneCommand("DEFAULT", Command.takeoff()),
                             SingleDroneCommand("DEFAULT", Command.up(300)),
                             SingleDroneCommand("DEFAULT", Command.down(100)),
                             SingleDroneCommand("DEFAULT", Command.land())]
        self.assertEqual(expected_commands, commands)

    def test_procedure_error_but_not_called_should_not_give_error(self):
        commands = generate_commands("""
            procedure error() {
              int a <- "error";
            }
            main () {
            }
            """)
        self.assertEqual([], commands)

    def test_call_not_defined_procedure_should_give_error(self):
        with self.assertRaises(CompileError) as context:
            generate_commands("""
                main () {
                  proc(100, 200);
                }
                """)
        self.assertTrue("Function or procedure proc has not been defined" in str(context.exception))

    def test_call_procedure_with_wrong_number_argument_should_give_error(self):
        with self.assertRaises(CompileError) as context:
            generate_commands("""
                procedure proc(int i, int j) {
                  up(i + j);
                  down(i);
                }
                main () {
                  takeoff();
                  proc(100);
                  land();
                }
                """)
        self.assertTrue("Arguments when calling function or procedure proc should have types {}, but is {}"
                        .format("['int', 'int']", "['int']")
                        in str(context.exception))

    def test_call_procedure_with_wrong_type_argument_should_give_error(self):
        types = [Type.int(), Type.decimal(), Type.string(), Type.boolean(), Type.vector(), Type.drone(),
                 Type.list_of(Type.int()), Type.list_of(Type.decimal()), Type.list_of(Type.list_of(Type.int()))]
        for t1 in types:
            for t2 in types:
                if t1 == t2:
                    continue
                with self.assertRaises(CompileError) as context:
                    generate_commands("""
                        procedure proc({} i) {{
                          down(i);
                        }}
                        main () {{
                          takeoff();
                          {} a;
                          proc(a);
                          land();
                        }}
                        """.format(t1, t2))
                self.assertTrue("Arguments when calling function or procedure proc should have types ['{}'], "
                                "but is ['{}']".format(t1, t2)
                                in str(context.exception))

    def test_call_procedure_but_ident_is_function_should_give_error(self):
        with self.assertRaises(CompileError) as context:
            generate_commands("""
                function func() return int {
                  return 1;
                }
                main () {
                  func();
                }
                """)
        self.assertTrue("Procedure call should not return any expression, but {} is returned"
                        .format(Expression(Type.int(), 1, ident=None))
                        in str(context.exception))

    def test_call_procedure_with_parameter_should_shadow_constant(self):
        commands = generate_commands("""
            procedure proc(drone DRONE2) {
              DRONE2.up(1);
            }
            main () {
              DRONE1.takeoff();
              proc(DRONE1);
              DRONE1.land();
            }
            """, drone_config_map={"DRONE1": DefaultDroneConfig(), "DRONE2": DefaultDroneConfig()})
        expected_commands = [SingleDroneCommand("DRONE1", Command.takeoff()),
                             SingleDroneCommand("DRONE1", Command.up(1)),
                             SingleDroneCommand("DRONE1", Command.land())]
        self.assertEqual(expected_commands, commands)


class FunctionCallTest(unittest.TestCase):
    def test_call_function_should_run_commands_and_return(self):
        commands = generate_commands("""
            function func(int i, int j) return int {
              up(i + j);
              down(i);
              return i + j;
            }
            main () {
              takeoff();
              forward(func(100, 200));
              land();
            }
            """)
        expected_commands = [SingleDroneCommand("DEFAULT", Command.takeoff()),
                             SingleDroneCommand("DEFAULT", Command.up(300)),
                             SingleDroneCommand("DEFAULT", Command.down(100)),
                             SingleDroneCommand("DEFAULT", Command.forward(300)),
                             SingleDroneCommand("DEFAULT", Command.land())]
        self.assertEqual(expected_commands, commands)

    def test_function_error_but_not_called_should_not_give_error(self):
        commands = generate_commands("""
            function func() return int {
              int a <- "error";
            }
            main () {
            }
            """)
        self.assertEqual([], commands)

    def test_call_not_defined_function_should_give_error(self):
        with self.assertRaises(CompileError) as context:
            generate_commands("""
                main () {
                  func(100, 200);
                }
                """)
        self.assertTrue("Function or procedure func has not been defined" in str(context.exception))

    def test_call_function_with_wrong_number_argument_should_give_error(self):
        with self.assertRaises(CompileError) as context:
            generate_commands("""
                function func(int i, int j) return int {
                  return i + j;
                }
                main () {
                  takeoff();
                  forward(func(100));
                  land();
                }
                """)
        self.assertTrue("Arguments when calling function or procedure func should have types {}, but is {}"
                        .format("['int', 'int']", "['int']")
                        in str(context.exception))

    def test_call_function_with_wrong_type_argument_should_give_error(self):
        types = [Type.int(), Type.decimal(), Type.string(), Type.boolean(), Type.vector(), Type.drone(),
                 Type.list_of(Type.int()), Type.list_of(Type.decimal()), Type.list_of(Type.list_of(Type.int()))]
        for t1 in types:
            for t2 in types:
                if t1 == t2:
                    continue
                with self.assertRaises(CompileError) as context:
                    generate_commands("""
                        function func({} i) return {} {{
                          return(i);
                        }}
                        main () {{
                          {} a;
                          {} b <- func(a);
                        }}
                        """.format(t1, t1, t2, t1))
                self.assertTrue("Arguments when calling function or procedure func should have types ['{}'], "
                                "but is ['{}']".format(t1, t2)
                                in str(context.exception))

    def test_call_function_but_ident_is_procedure_should_give_error(self):
        with self.assertRaises(CompileError) as context:
            generate_commands("""
                procedure proc() {
                }
                main () {
                  int a <- proc();
                }
                """)
        self.assertTrue("Function call should return an expression, but nothing is returned" in str(context.exception))

    def test_call_function_with_parameter_should_shadow_constant(self):
        commands = generate_commands("""
            function func(drone DRONE2) return int {
              DRONE2.up(1);
              return 1;
            }
            main () {
              DRONE1.takeoff();
              DRONE1.up(func(DRONE1));
              DRONE1.land();
            }
            """, drone_config_map={"DRONE1": DefaultDroneConfig(), "DRONE2": DefaultDroneConfig()})
        expected_commands = [SingleDroneCommand("DRONE1", Command.takeoff()),
                             SingleDroneCommand("DRONE1", Command.up(1)),
                             SingleDroneCommand("DRONE1", Command.up(1)),
                             SingleDroneCommand("DRONE1", Command.land())]
        self.assertEqual(expected_commands, commands)


class ComplexFunctionTest(unittest.TestCase):
    def test_recursion(self):
        commands = generate_commands("""
            function func(int i) return int {
              if i >= 10 {
                return 10;
              }
              return func(i + 1);
            }
            main () {
              takeoff();
              forward(func(1));
              land();
            }
            """)
        expected_commands = [SingleDroneCommand("DEFAULT", Command.takeoff()),
                             SingleDroneCommand("DEFAULT", Command.forward(10)),
                             SingleDroneCommand("DEFAULT", Command.land())]
        self.assertEqual(expected_commands, commands)

    def test_loops_in_function(self):
        commands = generate_commands("""
            function func(int i) return int {
              while i < 10 {
                i <- i + 1;
              }
              return i;
            }
            main () {
              takeoff();
              forward(func(1));
              land();
            }
            """)
        expected_commands = [SingleDroneCommand("DEFAULT", Command.takeoff()),
                             SingleDroneCommand("DEFAULT", Command.forward(10)),
                             SingleDroneCommand("DEFAULT", Command.land())]
        self.assertEqual(expected_commands, commands)

    def test_function_procedure_in_function(self):
        commands = generate_commands("""
            function func(int i) return int {
              while i < 10 {
                i <- i + 1;
              }
              return i;
            }
            procedure proc(int i) {
              forward(i * i);
            }
            function func2(int i) return int {
              proc(func(i));
              return func(i) * func(i);
            }
            main () {
              takeoff();
              forward(func2(1));
              land();
            }
            """)
        expected_commands = [SingleDroneCommand("DEFAULT", Command.takeoff()),
                             SingleDroneCommand("DEFAULT", Command.forward(100)),
                             SingleDroneCommand("DEFAULT", Command.forward(100)),
                             SingleDroneCommand("DEFAULT", Command.land())]
        self.assertEqual(expected_commands, commands)

    def test_scope(self):
        actual = SymbolTable()
        commands = generate_commands("""
            function func(int i) return int {
              int j <- 100;
              while i < 10 {
                i <- i + 1;
              }
              return i;
            }
            main () {
              takeoff();
              int i <- 1;
              forward(func(i));
              land();
            }
            """, symbol_table=actual)
        expected_commands = [SingleDroneCommand("DEFAULT", Command.takeoff()),
                             SingleDroneCommand("DEFAULT", Command.forward(10)),
                             SingleDroneCommand("DEFAULT", Command.land())]
        self.assertEqual(expected_commands, commands)
        expected = SymbolTable()
        expected.store("i", Expression(Type.int(), 1, ident="i"))
        self.assertEqual(expected, actual)


class ComplexProcedureTest(unittest.TestCase):
    def test_recursion(self):
        commands = generate_commands("""
            procedure proc(int i) {
              if i >= 10 {
                forward(10);
                return;
              }
              proc(i + 1);
            }
            main () {
              takeoff();
              proc(1);
              land();
            }
            """)
        expected_commands = [SingleDroneCommand("DEFAULT", Command.takeoff()),
                             SingleDroneCommand("DEFAULT", Command.forward(10)),
                             SingleDroneCommand("DEFAULT", Command.land())]
        self.assertEqual(expected_commands, commands)

    def test_loops_in_function(self):
        commands = generate_commands("""
            procedure proc(int i) {
              while i < 10 {
                i <- i + 1;
              }
              forward(i);
            }
            main () {
              takeoff();
              proc(1);
              land();
            }
            """)
        expected_commands = [SingleDroneCommand("DEFAULT", Command.takeoff()),
                             SingleDroneCommand("DEFAULT", Command.forward(10)),
                             SingleDroneCommand("DEFAULT", Command.land())]
        self.assertEqual(expected_commands, commands)

    def test_function_procedure_in_procedure(self):
        commands = generate_commands("""
            function func(int i) return int {
              return i * i;
            }
            procedure proc(int i) {
              forward(i);
            }
            procedure proc2() {
              takeoff();
              proc(func(10));
              proc(func(10));
              land();
            }
            main () {
              proc2();
            }
            """)
        expected_commands = [SingleDroneCommand("DEFAULT", Command.takeoff()),
                             SingleDroneCommand("DEFAULT", Command.forward(100)),
                             SingleDroneCommand("DEFAULT", Command.forward(100)),
                             SingleDroneCommand("DEFAULT", Command.land())]
        self.assertEqual(expected_commands, commands)

    def test_scope(self):
        actual = SymbolTable()
        generate_commands("""
            procedure proc(int i) {
              int j <- 100;
              while i < 10 {
                i <- i + 1;
              }
            }
            main () {
              int i <- 1;
              proc(i);
            }
            """, symbol_table=actual)
        expected = SymbolTable()
        expected.store("i", Expression(Type.int(), 1, ident="i"))
        self.assertEqual(expected, actual)



import unittest
from math import sqrt

from xdrone.shared.command import Command
from xdrone.shared.drone_config import DroneConfig
from xdrone.shared.state import State
from xdrone.state_updaters.state_updater import StateUpdater


class StateUpdaterTest(unittest.TestCase):
    def setUp(self) -> None:
        self.drone_config = DroneConfig(init_position=(0, 0, 0), speed_mps=0.5,
                                        rotate_speed_dps=45, takeoff_height_meters=1)
        self.state_updater = StateUpdater(drone_config=self.drone_config)

    def test_get_init_state(self):
        self.assertEqual(State(x_meters=0, y_meters=0, z_meters=0), self.state_updater.get_init_state())
        drone_config = DroneConfig(init_position=(1, 2, 3), speed_mps=0.5,
                                   rotate_speed_dps=45, takeoff_height_meters=1)
        state_updater = StateUpdater(drone_config=drone_config)
        self.assertEqual(State(x_meters=1, y_meters=2, z_meters=3), state_updater.get_init_state())

    def test_update_takeoff_should_update_state(self):
        state = State()
        actual = self.state_updater.update(Command.takeoff(), state)
        expected = State(has_taken_off=True, time_used_seconds=self.drone_config.takeoff_height_meters / 0.5,
                         z_meters=self.drone_config.takeoff_height_meters)
        self.assertEqual(expected, actual)

    def test_update_land_should_update_state(self):
        state = State(has_taken_off=True, z_meters=3)
        actual = self.state_updater.update(Command.land(), state)
        expected = State(has_taken_off=False, time_used_seconds=3 / 0.5, z_meters=0)
        self.assertEqual(expected, actual)

    def test_update_up_should_update_state(self):
        state = State(has_taken_off=True, z_meters=3)
        actual = self.state_updater.update(Command.up(1), state)
        expected = State(has_taken_off=True, time_used_seconds=1 / 0.5, z_meters=3 + 1)
        self.assertEqual(expected, actual)

    def test_update_down_should_update_state(self):
        state = State(has_taken_off=True, z_meters=3)
        actual = self.state_updater.update(Command.down(1), state)
        expected = State(has_taken_off=True, time_used_seconds=1 / 0.5, z_meters=3 - 1)
        self.assertEqual(expected, actual)

    def test_update_left_should_update_state(self):
        state = State(has_taken_off=True, x_meters=1, y_meters=2, z_meters=3, orientation_degrees=60)
        actual = self.state_updater.update(Command.left(1), state)
        expected = State(has_taken_off=True, time_used_seconds=1 / 0.5, x_meters=1 - 0.5, y_meters=2 + sqrt(3) / 2,
                         z_meters=3, orientation_degrees=60)
        self.assertEqual(expected, actual)

        state = State(has_taken_off=True, x_meters=1, y_meters=2, z_meters=3, orientation_degrees=240)
        actual = self.state_updater.update(Command.left(1), state)
        expected = State(has_taken_off=True, time_used_seconds=1 / 0.5, x_meters=1 + 0.5, y_meters=2 - sqrt(3) / 2,
                         z_meters=3, orientation_degrees=240)
        self.assertEqual(expected, actual)

    def test_update_right_should_update_state(self):
        state = State(has_taken_off=True, x_meters=1, y_meters=2, z_meters=3, orientation_degrees=60)
        actual = self.state_updater.update(Command.right(1), state)
        expected = State(has_taken_off=True, time_used_seconds=1 / 0.5, x_meters=1 + 0.5, y_meters=2 - sqrt(3) / 2,
                         z_meters=3, orientation_degrees=60)
        self.assertEqual(expected, actual)

        state = State(has_taken_off=True, x_meters=1, y_meters=2, z_meters=3, orientation_degrees=240)
        actual = self.state_updater.update(Command.right(1), state)
        expected = State(has_taken_off=True, time_used_seconds=1 / 0.5, x_meters=1 - 0.5, y_meters=2 + sqrt(3) / 2,
                         z_meters=3, orientation_degrees=240)
        self.assertEqual(expected, actual)

    def test_update_forward_should_update_state(self):
        state = State(has_taken_off=True, x_meters=1, y_meters=2, z_meters=3, orientation_degrees=60)
        actual = self.state_updater.update(Command.forward(1), state)
        expected = State(has_taken_off=True, time_used_seconds=1 / 0.5, x_meters=1 + sqrt(3) / 2, y_meters=2 + 0.5,
                         z_meters=3, orientation_degrees=60)
        self.assertEqual(expected, actual)

        state = State(has_taken_off=True, x_meters=1, y_meters=2, z_meters=3, orientation_degrees=240)
        actual = self.state_updater.update(Command.forward(1), state)
        expected = State(has_taken_off=True, time_used_seconds=1 / 0.5, x_meters=1 - sqrt(3) / 2, y_meters=2 - 0.5,
                         z_meters=3, orientation_degrees=240)
        self.assertEqual(expected, actual)

    def test_update_backward_should_update_state(self):
        state = State(has_taken_off=True, x_meters=1, y_meters=2, z_meters=3, orientation_degrees=60)
        actual = self.state_updater.update(Command.backward(1), state)
        expected = State(has_taken_off=True, time_used_seconds=1 / 0.5, x_meters=1 - sqrt(3) / 2, y_meters=2 - 0.5,
                         z_meters=3, orientation_degrees=60)
        self.assertEqual(expected, actual)

        state = State(has_taken_off=True, x_meters=1, y_meters=2, z_meters=3, orientation_degrees=240)
        actual = self.state_updater.update(Command.backward(1), state)
        expected = State(has_taken_off=True, time_used_seconds=1 / 0.5, x_meters=1 + sqrt(3) / 2, y_meters=2 + 0.5,
                         z_meters=3, orientation_degrees=240)
        self.assertEqual(expected, actual)

    def test_update_rotate_left_should_update_state(self):
        state = State(has_taken_off=True, x_meters=1, y_meters=2, z_meters=3, orientation_degrees=60)
        actual = self.state_updater.update(Command.rotate_left(90), state)
        expected = State(has_taken_off=True, time_used_seconds=1 / 0.5, x_meters=1, y_meters=2,
                         z_meters=3, orientation_degrees=330)
        self.assertEqual(expected, actual)

        state = State(has_taken_off=True, x_meters=1, y_meters=2, z_meters=3, orientation_degrees=240)
        actual = self.state_updater.update(Command.rotate_left(90), state)
        expected = State(has_taken_off=True, time_used_seconds=1 / 0.5, x_meters=1, y_meters=2,
                         z_meters=3, orientation_degrees=150)
        self.assertEqual(expected, actual)

    def test_update_rotate_right_should_update_state(self):
        state = State(has_taken_off=True, x_meters=1, y_meters=2, z_meters=3, orientation_degrees=60)
        actual = self.state_updater.update(Command.rotate_right(90), state)
        expected = State(has_taken_off=True, time_used_seconds=1 / 0.5, x_meters=1, y_meters=2,
                         z_meters=3, orientation_degrees=150)
        self.assertEqual(expected, actual)

        state = State(has_taken_off=True, x_meters=1, y_meters=2, z_meters=3, orientation_degrees=240)
        actual = self.state_updater.update(Command.rotate_right(90), state)
        expected = State(has_taken_off=True, time_used_seconds=1 / 0.5, x_meters=1, y_meters=2,
                         z_meters=3, orientation_degrees=330)
        self.assertEqual(expected, actual)

    def test_update_wait_should_update_state(self):
        state = State(has_taken_off=True, x_meters=1, y_meters=2, z_meters=3, orientation_degrees=60)
        actual = self.state_updater.update(Command.wait(5), state)
        expected = State(has_taken_off=True, time_used_seconds=5, x_meters=1, y_meters=2,
                         z_meters=3, orientation_degrees=60)
        self.assertEqual(expected, actual)



from typing import List, Dict, Set

from xdrone.shared.boundary_config import BoundaryConfig
from xdrone.shared.command import AbstractDroneCommand, SingleDroneCommand, ParallelDroneCommands
from xdrone.shared.safety_check_error import SafetyCheckError
from xdrone.shared.state import State
from xdrone.state_updaters.state_updater import StateUpdater


class BoundaryChecker:
    def __init__(self, boundary_config: BoundaryConfig):
        self.boundary_config = boundary_config

    def check(self, drone_commands: List[AbstractDroneCommand], state_updater_map: Dict[str, StateUpdater]):
        drone_state_map = {name: state_updater.get_init_state() for name, state_updater in state_updater_map.items()}
        drones_involved = set(state_updater_map.keys())
        self._update_states_and_check_drone_commands(drone_commands, state_updater_map, drone_state_map,
                                                     drones_involved)
        for name, state in drone_state_map.items():
            if state.has_taken_off:
                raise SafetyCheckError("Drone '{}' did not land in the end".format(name))

    def _update_states_and_check_drone_commands(self, drone_commands: List[AbstractDroneCommand],
                                                state_updaters: Dict[str, StateUpdater],
                                                drone_state_map: Dict[str, State],
                                                drones_involved: Set[str]) -> float:
        total_time_used = 0
        for drone_command in drone_commands:
            if isinstance(drone_command, SingleDroneCommand):
                total_time_used += self._update_states_and_check_single_drone_command(drone_command,
                                                                                      state_updaters,
                                                                                      drone_state_map,
                                                                                      drones_involved)
            elif isinstance(drone_command, ParallelDroneCommands):
                total_time_used += self._update_states_and_check_parallel_drone_commands(drone_command,
                                                                                         state_updaters,
                                                                                         drone_state_map,
                                                                                         drones_involved)
        return total_time_used

    def _update_states_and_check_single_drone_command(self, single_drone_command: SingleDroneCommand,
                                                      state_updaters: Dict[str, StateUpdater],
                                                      drone_state_map: Dict[str, State],
                                                      drones_involved: Set[str]) -> float:
        self._check_takeoff_state(single_drone_command, drone_state_map)
        time_used = self._update_states(single_drone_command, state_updaters, drone_state_map, drones_involved)
        for name, state in drone_state_map.items():
            try:
                self.boundary_config.check_state(name, state)
            except SafetyCheckError as e:
                raise SafetyCheckError("When running command '{}', boundary limits are violated: {}"
                                       .format(single_drone_command.to_command_str(), str(e)))
        return time_used

    def _check_takeoff_state(self, single_drone_command: SingleDroneCommand, drone_state_map: Dict[str, State]):
        drone_name = single_drone_command.drone_name
        command = single_drone_command.command
        if command.opcode == 'wait':
            # `wait` can be called in both cases
            pass
        elif command.opcode == "takeoff":
            # `takeoff` can only be called when has not taken off
            if drone_state_map[drone_name].has_taken_off:
                raise SafetyCheckError("'{}' command used when drone '{}' has already been taken off"
                                       .format(command.opcode, drone_name))
        else:
            # other commands can only be called when has taken off
            if not drone_state_map[drone_name].has_taken_off:
                raise SafetyCheckError("'{}' command used when drone '{}' has not been taken off"
                                       .format(command.opcode, drone_name))

    def _update_states(self, single_drone_command: SingleDroneCommand,
                       state_updaters: Dict[str, StateUpdater],
                       drone_state_map: Dict[str, State],
                       drones_involved: Set[str]) -> float:
        drone_name = single_drone_command.drone_name
        command = single_drone_command.command
        old_state = drone_state_map[drone_name]
        new_state = state_updaters[drone_name].update(command, old_state)
        drone_state_map[drone_name] = new_state
        # update other drones
        time_used = new_state.time_used_seconds - old_state.time_used_seconds
        for name in drones_involved.difference({drone_name}):
            state = drone_state_map[name]
            drone_state_map[name] = state.copy_and_set_time_used_seconds(state.time_used_seconds + time_used)
        return time_used

    def _update_states_and_check_parallel_drone_commands(self, parallel_drone_commands: ParallelDroneCommands,
                                                         state_updaters: Dict[str, StateUpdater],
                                                         drone_state_map: Dict[str, State],
                                                         drones_involved: Set[str]) -> float:
        assert len(parallel_drone_commands.branches) > 0
        time_used_in_branches = self._update_states_and_check_for_each_branch(parallel_drone_commands,
                                                                              state_updaters,
                                                                              drone_state_map)
        longest_time_used = self._update_states_to_wait_for_slowest_branch(parallel_drone_commands,
                                                                           drone_state_map,
                                                                           time_used_in_branches,
                                                                           drones_involved)

        for name, state in drone_state_map.items():
            try:
                self.boundary_config.check_state(name, state)
            except SafetyCheckError as e:
                raise SafetyCheckError("When running command '{}', boundary limits are violated: {}"
                                       .format(parallel_drone_commands.to_command_str(), str(e)))
        return longest_time_used

    def _update_states_and_check_for_each_branch(self, parallel_drone_commands: ParallelDroneCommands,
                                                 state_updaters: Dict[str, StateUpdater],
                                                 drone_state_map: Dict[str, State]) -> List[float]:
        time_used_in_branches = []
        for i, branch in enumerate(parallel_drone_commands.branches):
            drones_involved = parallel_drone_commands.drones_involved_each_branch[i]
            time_used = self._update_states_and_check_drone_commands(branch, state_updaters,
                                                                     drone_state_map, drones_involved)
            time_used_in_branches.append(time_used)
        return time_used_in_branches

    def _update_states_to_wait_for_slowest_branch(self, parallel_drone_commands: ParallelDroneCommands,
                                                  drone_state_map: Dict[str, State],
                                                  time_used_in_branches: List[float],
                                                  drones_involved: Set[str]) -> float:
        longest_time_used = max(time_used_in_branches)
        # for each branch, let drones involved in the branch wait until longest_time_used
        for i, time_used in enumerate(time_used_in_branches):
            for name in parallel_drone_commands.drones_involved_each_branch[i]:
                old_state = drone_state_map[name]
                drone_state_map[name] = old_state.copy_and_set_time_used_seconds(
                    old_state.time_used_seconds + longest_time_used - time_used)
        # let drones not involved in any branch wait for longest_time_used
        for name in drones_involved.difference(parallel_drone_commands.get_drones_involved()):
            old_state = drone_state_map[name]
            drone_state_map[name] = old_state.copy_and_set_time_used_seconds(
                old_state.time_used_seconds + longest_time_used)
        return longest_time_used



from math import sin, cos, radians

from xdrone.shared.command import Command
from xdrone.shared.drone_config import DroneConfig
from xdrone.shared.state import State


class StateUpdater:
    def __init__(self, drone_config: DroneConfig):
        self.drone_config = drone_config
        self.OPCODE_METHOD_MAP = {
            "takeoff": self._update_takeoff,
            "land": self._update_land,
            "up": self._update_up,
            "down": self._update_down,
            "left": self._update_left,
            "right": self._update_right,
            "forward": self._update_forward,
            "backward": self._update_backward,
            "rotate_left": self._update_rotate_left,
            "rotate_right": self._update_rotate_right,
            "wait": self._update_wait
        }

    def get_init_state(self) -> State:
        (x, y, z) = self.drone_config.init_position
        return State(x_meters=x, y_meters=y, z_meters=z)

    def update(self, command: Command, state: State) -> State:
        update_method = self.OPCODE_METHOD_MAP[command.opcode]
        return update_method(command, state)

    def _update_takeoff(self, command: Command, state: State) -> State:
        assert command.opcode == "takeoff"
        state = state.copy_and_set_has_taken_off(True)
        state = state.copy_and_set_time_used_seconds(
            state.time_used_seconds + self.drone_config.takeoff_height_meters / self.drone_config.speed_mps)
        state = state.copy_and_set_z_meters(self.drone_config.takeoff_height_meters)
        return state

    def _update_land(self, command: Command, state: State) -> State:
        assert command.opcode == "land"
        state = state.copy_and_set_has_taken_off(False)
        state = state.copy_and_set_time_used_seconds(
            state.time_used_seconds + state.z_meters / self.drone_config.speed_mps)
        state = state.copy_and_set_z_meters(0)
        return state

    def _update_up(self, command: Command, state: State) -> State:
        assert command.opcode == "up"
        up_meters, = command.operands
        state = state.copy_and_set_time_used_seconds(state.time_used_seconds + up_meters / self.drone_config.speed_mps)
        state = state.copy_and_set_z_meters(state.z_meters + up_meters)
        return state

    def _update_down(self, command: Command, state: State) -> State:
        assert command.opcode == "down"
        down_meters, = command.operands
        state = state.copy_and_set_time_used_seconds(
            state.time_used_seconds + down_meters / self.drone_config.speed_mps)
        state = state.copy_and_set_z_meters(state.z_meters - down_meters)
        return state

    def _update_left(self, command: Command, state: State) -> State:
        assert command.opcode == "left"
        left_meters, = command.operands
        move_direction = state.orientation_degrees - 90
        state = state.copy_and_set_time_used_seconds(
            state.time_used_seconds + left_meters / self.drone_config.speed_mps)
        state = state.copy_and_set_x_meters(state.x_meters + left_meters * sin(radians(move_direction)))
        state = state.copy_and_set_y_meters(state.y_meters + left_meters * cos(radians(move_direction)))
        return state

    def _update_right(self, command: Command, state: State) -> State:
        assert command.opcode == "right"
        right_meters, = command.operands
        move_direction = state.orientation_degrees + 90
        state = state.copy_and_set_time_used_seconds(
            state.time_used_seconds + right_meters / self.drone_config.speed_mps)
        state = state.copy_and_set_x_meters(state.x_meters + right_meters * sin(radians(move_direction)))
        state = state.copy_and_set_y_meters(state.y_meters + right_meters * cos(radians(move_direction)))
        return state

    def _update_forward(self, command: Command, state: State) -> State:
        assert command.opcode == "forward"
        forward_meters, = command.operands
        move_direction = state.orientation_degrees
        state = state.copy_and_set_time_used_seconds(
            state.time_used_seconds + forward_meters / self.drone_config.speed_mps)
        state = state.copy_and_set_x_meters(state.x_meters + forward_meters * sin(radians(move_direction)))
        state = state.copy_and_set_y_meters(state.y_meters + forward_meters * cos(radians(move_direction)))
        return state

    def _update_backward(self, command: Command, state: State) -> State:
        assert command.opcode == "backward"
        backward_meters, = command.operands
        move_direction = state.orientation_degrees + 180
        state = state.copy_and_set_time_used_seconds(
            state.time_used_seconds + backward_meters / self.drone_config.speed_mps)
        state = state.copy_and_set_x_meters(state.x_meters + backward_meters * sin(radians(move_direction)))
        state = state.copy_and_set_y_meters(state.y_meters + backward_meters * cos(radians(move_direction)))
        return state

    def _update_rotate_left(self, command: Command, state: State) -> State:
        assert command.opcode == "rotate_left"
        rotate_left_degrees, = command.operands
        state = state.copy_and_set_time_used_seconds(
            state.time_used_seconds + rotate_left_degrees / self.drone_config.rotate_speed_dps)
        state = state.copy_and_set_orientation_degrees((state.orientation_degrees - rotate_left_degrees) % 360)
        return state

    def _update_rotate_right(self, command: Command, state: State) -> State:
        assert command.opcode == "rotate_right"
        rotate_right_degrees, = command.operands
        state = state.copy_and_set_time_used_seconds(
            state.time_used_seconds + rotate_right_degrees / self.drone_config.rotate_speed_dps)
        state = state.copy_and_set_orientation_degrees((state.orientation_degrees + rotate_right_degrees) % 360)
        return state

    def _update_wait(self, command: Command, state: State) -> State:
        assert command.opcode == "wait"
        wait_seconds, = command.operands
        state = state.copy_and_set_time_used_seconds(state.time_used_seconds + wait_seconds)
        return state



import socket
import threading
import time
from logging import info, error
from typing import List

from xdrone.shared.command import Command


class DJITelloExecutor:
    TELLO_IP = "192.168.10.1"
    COMMAND_PORT = 8889
    HOST_IP = "0.0.0.0"
    RESPONSE_PORT = 9000

    def __init__(self):
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.sock.bind((DJITelloExecutor.HOST_IP, DJITelloExecutor.RESPONSE_PORT))
        self.recv_thread = threading.Thread(name="Thread-Receive", target=self._recv)
        self.recv_thread.daemon = True
        self.running = True
        self.waiting_for_resp = False

    def _terminate(self):
        while self.waiting_for_resp:
            pass
        self.running = False
        self.sock.close()

    def _recv(self):
        while True:
            while not self.waiting_for_resp:
                if not self.running:
                    return
            try:
                msg, _ = self.sock.recvfrom(1024)
                info("received response: {}".format(msg.decode(encoding="utf-8")))
            except Exception as err:
                if self.running:
                    error("Error met when receiving response: " + str(err))
            self.waiting_for_resp = False

    def _send(self, msg):
        while self.waiting_for_resp:
            pass
        msg = msg.encode(encoding="utf-8")
        self.sock.sendto(msg, (DJITelloExecutor.TELLO_IP, DJITelloExecutor.COMMAND_PORT))
        info("sent message: {}".format(msg))
        self.waiting_for_resp = True

    def _emergency(self):
        msg = "emergency".encode(encoding="utf-8")
        self.sock.sendto(msg, (DJITelloExecutor.TELLO_IP, DJITelloExecutor.COMMAND_PORT))
        self.waiting_for_resp = False

    def execute_commands(self, commands: List[Command]):
        try:
            self.recv_thread.start()

            self._send("command")
            for command in commands:
                if command.opcode == "takeoff":
                    self._send("takeoff")
                if command.opcode == "land":
                    self._send("land")
                if command.opcode == "up":
                    self._send("up" + " " + str(int(command.operands[0] * 100)))
                if command.opcode == "down":
                    self._send("down" + " " + str(int(command.operands[0] * 100)))
                if command.opcode == "left":
                    self._send("left" + " " + str(int(command.operands[0] * 100)))
                if command.opcode == "right":
                    self._send("right" + " " + str(int(command.operands[0] * 100)))
                if command.opcode == "forward":
                    self._send("forward" + " " + str(int(command.operands[0] * 100)))
                if command.opcode == "backward":
                    self._send("back" + " " + str(int(command.operands[0] * 100)))
                if command.opcode == "rotate_left":
                    self._send("ccw" + " " + str(int(command.operands[0])))
                if command.opcode == "rotate_right":
                    self._send("cw" + " " + str(int(command.operands[0])))
                if command.opcode == "wait":
                    time.sleep(command.operands[0])
        except KeyboardInterrupt:
            self._emergency()
            info("KeyboardInterrupt received. Forced stop.")
        finally:
            self._terminate()
            if self.recv_thread.is_alive():
                self.recv_thread.join()



import copy
from typing import List, Optional

from xdrone.compiler.compiler_utils.type import Type


class FunctionIdentifier:
    def __init__(self, ident: str):
        self._ident = ident

    @property
    def ident(self) -> str:
        return copy.deepcopy(self._ident)

    def __str__(self):
        return "FunctionIdentifier: {{ ident: {} }}".format(self._ident)

    def __eq__(self, other):
        if isinstance(other, FunctionIdentifier):
            return other._ident == self._ident
        return False


class Parameter:
    def __init__(self, ident: str, type: Type):
        self._ident = ident
        self._type = type

    @property
    def ident(self) -> str:
        return copy.deepcopy(self._ident)

    @property
    def type(self) -> Type:
        return copy.deepcopy(self._type)

    def __str__(self):
        return "Parameter: {{ ident: {}, type: {} }}".format(self._ident, self._type)

    def __eq__(self, other):
        if isinstance(other, Parameter):
            return other._ident == self._ident and other._type == self._type
        return False


class Function:
    """
    Note that the attribute `commands` is a list of antlr contexts, which is difficult to compare and copy,
    so we do NOT regard the attribute `commands` as a property of a Function object,
    e.g. in __eq__ the attribute `commands` will not be compared,
    in __str__ the attribute `commands` will not be printed
    """

    def __init__(self, ident: str, param_list: List[Parameter], return_type: Optional[Type], commands: list):
        self._ident = ident
        self._param_list = param_list
        self._return_type = return_type
        self._commands = commands

    @property
    def ident(self) -> str:
        return copy.deepcopy(self._ident)

    @property
    def param_list(self) -> List[Parameter]:
        return copy.deepcopy(self._param_list)

    @property
    def return_type(self) -> Type:
        return copy.deepcopy(self._return_type)

    def get_commands(self) -> list:
        # deepcopy will give error, so use shallow copy
        return self._commands

    def __str__(self):
        param_list_str = "[" + ", ".join(map(str, self._param_list)) + "]"
        return "Function: {{ ident: {}, param_list: {}, return_type: {} }}" \
            .format(self._ident, param_list_str, self._return_type)

    def __eq__(self, other):
        if isinstance(other, Function):
            return other._ident == self._ident and other._param_list == self._param_list \
                   and other._return_type == self._return_type
        return False


class FunctionTable:
    def __init__(self):
        self._function_table = {}

    def __str__(self):
        result = "FunctionTable: {\n"
        for ident, function in self._function_table.items():
            result += "  " + ident + " -> " + str(function) + "\n"
        result += "}"
        return result

    def __eq__(self, other):
        if isinstance(other, FunctionTable):
            return other._function_table == self._function_table
        return False

    def __contains__(self, ident: str) -> bool:
        return ident in self._function_table

    def store(self, ident: str, function: Function) -> None:
        assert not self.__contains__(ident), "Function already in function table"
        self._function_table[ident] = function

    def get_function(self, ident: str) -> Function:
        assert self.__contains__(ident), "Function not in symbol table"
        # deepcopy will give error due to `commands` in Function, so use shallow copy
        return self._function_table[ident]
